"""
Script to convert the EchoNet-LVH database to zea format.

Each video is cropped so that the scan cone is centered
without padding, such that it can be converted to polar domain.

This cropping requires first computing scan cone parameters
using `data/convert/echonetlvh/precompute_crop.py`, which
are then passed to this script.
"""

import os

os.environ["KERAS_BACKEND"] = "jax"


if __name__ == "__main__":
    from zea import init_device

    init_device("auto:1")

import argparse
import csv
import sys
from concurrent.futures import ProcessPoolExecutor, as_completed
from pathlib import Path

import jax.numpy as jnp
import numpy as np
from jax import jit, vmap
from tqdm import tqdm

from zea.data import generate_zea_dataset
from zea.data.convert.echonet import H5Processor
from zea.display import cartesian_to_polar_matrix
from zea.io_lib import load_video
from zea.utils import translate


def get_args():
    """Parse command line arguments."""
    parser = argparse.ArgumentParser(description="Convert EchoNet-LVH to zea format")
    parser.add_argument(
        "--source",
        type=str,
        required=True,
    )
    parser.add_argument("--output", type=str, required=True)
    parser.add_argument("--output_numpy", type=str, default=None)
    parser.add_argument("--file_list", type=str, help="Optional path to list of files")
    parser.add_argument("--use_hyperthreading", action="store_true", help="Enable hyperthreading")
    parser.add_argument(
        "--batch",
        type=str,
        help="Specify which BatchX directory to process, e.g. --batch=Batch2",
    )
    parser.add_argument(
        "--max_files",
        type=int,
        default=None,
        help="Maximum number of files to process (for testing)",
    )
    # if neither is specified, both will be converted
    parser.add_argument(
        "--convert_measurements",
        action="store_true",
        help="Only convert measurements CSV file",
    )
    parser.add_argument("--convert_images", action="store_true", help="Only convert image files")
    return parser.parse_args()


def load_splits(source_dir):
    """Load splits from MeasurementsList.csv and return avi filenames"""
    csv_path = Path(source_dir) / "MeasurementsList.csv"
    splits = {"train": [], "val": [], "test": []}
    with open(csv_path, newline="", encoding="utf-8") as csvfile:
        reader = csv.DictReader(csvfile)
        file_split_map = {}
        for row in reader:
            filename = row["HashedFileName"]
            split = row["split"]
            file_split_map.setdefault(filename, split)
        for filename, split in file_split_map.items():
            splits[split].append(filename + ".avi")
    return splits


def find_avi_file(source_dir, hashed_filename, batch=None):
    """Find AVI file in the specified batch directory or any batch if not specified."""
    # If filename already has .avi extension, strip it
    if hashed_filename.endswith(".avi"):
        hashed_filename = hashed_filename[:-4]

    if batch:
        batch_dir = Path(source_dir) / batch
        avi_path = batch_dir / f"{hashed_filename}.avi"
        if avi_path.exists():
            return avi_path
        return None
    else:
        for batch_dir in Path(source_dir).glob("Batch*"):
            avi_path = batch_dir / f"{hashed_filename}.avi"
            if avi_path.exists():
                return avi_path
        return None


def load_cone_parameters(csv_path):
    """
    Load cone parameters from CSV file into a dictionary.

    Args:
        csv_path: Path to the CSV file containing cone parameters

    Returns:
        Dictionary mapping avi_filename to cone parameters
    """
    cone_params = {}

    with open(csv_path, "r", encoding="utf-8") as csvfile:
        reader = csv.DictReader(csvfile)
        for row in reader:
            if row["status"] == "success":
                # Convert string values to appropriate types
                params = {}
                for key, value in row.items():
                    if key in ("avi_filename", "status"):
                        params[key] = value
                    elif key == "apex_above_image":
                        params[key] = value.lower() == "true"
                    elif value is not None and value != "":
                        params[key] = float(value)
                    else:
                        params[key] = None

                cone_params[row["avi_filename"]] = params

    return cone_params


def crop_frame_with_params(frame, cone_params):
    """
    Crop a single frame using predetermined cone parameters.

    Args:
        frame: Input frame as numpy array
        cone_params: Dictionary containing cropping parameters

    Returns:
        Cropped and padded frame
    """
    crop_left = int(cone_params["crop_left"])
    crop_right = int(cone_params["crop_right"])
    crop_top = int(cone_params["crop_top"])
    crop_bottom = int(cone_params["crop_bottom"])

    # Handle negative crop_top
    if crop_top < 0:
        cropped = frame[0:crop_bottom, crop_left:crop_right]
        # Add top padding
        top_padding = -crop_top
        top_pad = jnp.zeros((top_padding, cropped.shape[1]), dtype=cropped.dtype)
        cropped = jnp.concatenate([top_pad, cropped], axis=0)
    else:
        cropped = frame[crop_top:crop_bottom, crop_left:crop_right]

    # Apply horizontal centering
    apex_x_in_crop = cone_params["apex_x"] - crop_left
    cropped_height, cropped_width = cropped.shape
    target_center_x = cropped_width / 2
    left_padding_needed = target_center_x - apex_x_in_crop

    left_padding = max(0, int(left_padding_needed))
    right_padding = max(0, int(-left_padding_needed))

    if left_padding > 0 or right_padding > 0:
        if left_padding > 0:
            left_pad = jnp.zeros((cropped_height, left_padding), dtype=cropped.dtype)
            cropped = jnp.concatenate([left_pad, cropped], axis=1)

        if right_padding > 0:
            right_pad = jnp.zeros((cropped_height, right_padding), dtype=cropped.dtype)
            cropped = jnp.concatenate([cropped, right_pad], axis=1)

    return cropped


def crop_sequence_with_params(sequence, cone_params):
    """
    Apply cropping to a sequence of frames using predetermined parameters.

    Args:
        sequence: Input sequence as numpy array of shape (frames, height, width)
        cone_params: Dictionary containing cropping parameters

    Returns:
        Cropped and padded sequence
    """
    crop_sequence = vmap(lambda frame: crop_frame_with_params(frame, cone_params))
    return crop_sequence(sequence)


class LVHProcessor(H5Processor):
    """Modified H5Processor for EchoNet-LVH dataset."""

    def __init__(self, *args, cone_params=None, **kwargs):
        super().__init__(*args, **kwargs)
        # self.cart2pol_jit = jit(cartesian_to_polar_matrix_jax)
        self.cart2pol_jit = jit(cartesian_to_polar_matrix)
        self.cart2pol_batched = vmap(self.cart2pol_jit)
        # Store the pre-computed cone parameters
        self.cone_parameters = cone_params or {}

    def get_split(self, avi_file: str, sequence):
        """
        Get the split (train/val/test) for a given AVI file.

        Args:
            avi_file: Path to the AVI file
            sequence: Video sequence (unused)

        Returns:
            String indicating the split ('train', 'val', or 'test')
        """
        # Extract base filename without extension
        filename = Path(avi_file).stem + ".avi"

        for split, files in self.splits.items():
            if filename in files:
                return split
        raise UserWarning("Unknown split for file: " + filename)

    def __call__(self, avi_file):
        print(avi_file)
        avi_filename = Path(avi_file).stem + ".avi"
        sequence = jnp.array(load_video(avi_file))

        sequence = translate(sequence, self.range_from, self._process_range)

        # Get pre-computed cone parameters for this file
        cone_params = self.cone_parameters.get(avi_filename)

        if cone_params is not None:
            # Apply pre-computed cropping parameters
            sequence = crop_sequence_with_params(sequence, cone_params)
        else:
            print(f"Warning: No cone parameters for {avi_filename}, using original sequence")

        # Convert to JAX array for polar conversion
        sequence = jnp.array(sequence)

        split = self.get_split(avi_file, sequence)
        out_h5 = self.path_out_h5 / split / (Path(avi_file).stem + ".hdf5")

        polar_im_set = self.cart2pol_batched(sequence)

        zea_dataset = {
            "path": out_h5,
            # store as uint8 for memory efficiency
            "image_sc": translate(np.array(sequence), self._process_range, (0, 255)).astype(
                np.uint8
            ),
            "probe_name": "generic",
            "description": "EchoNet-LVH dataset converted to zea format",
            "image": translate(np.array(polar_im_set), self._process_range, (0, 255)).astype(
                np.uint8
            ),
            "cast_to_float": False,
        }
        return generate_zea_dataset(**zea_dataset)


def transform_measurement_coordinates_with_cone_params(row, cone_params):
    """Transform measurement coordinates using cone parameters from fit_scan_cone.

    Args:
        row: A dict containing measurement data with X1,X2,Y1,Y2 coordinates
        cone_params: Dictionary containing cone parameters from fit_scan_cone

    Returns:
        A new row with transformed coordinates, or None if cone_params is None
    """
    if cone_params is None:
        print(f"Warning: No cone parameters for file {row['HashedFileName']}")
        return None

    new_row = dict(row)

    # Apply cropping offset
    crop_left = cone_params["crop_left"]
    crop_top = cone_params["crop_top"]

    # Transform coordinates
    for k in ["X1", "X2", "Y1", "Y2"]:
        # Convert to float if not already
        new_row[k] = float(row[k]) - (crop_left if k.startswith("X") else crop_top)

    # Apply horizontal centering offset
    apex_x_in_crop = cone_params["apex_x"] - crop_left
    original_width = cone_params["crop_right"] - cone_params["crop_left"]
    target_center_x = original_width / 2
    left_padding_needed = target_center_x - apex_x_in_crop
    left_padding = max(0, int(left_padding_needed))

    # Adjust x coordinates for horizontal padding
    new_row["X1"] = new_row["X1"] + left_padding
    new_row["X2"] = new_row["X2"] + left_padding

    # Check if coordinates are within the final image bounds
    final_width = cone_params["new_width"]
    final_height = cone_params["new_height"]

    # Check if coordinates are out of bounds
    is_out_of_bounds = (
        new_row["X1"] < 0
        or new_row["X2"] < 0
        or new_row["Y1"] < 0
        or new_row["Y2"] < 0
        or new_row["X1"] >= final_width
        or new_row["X2"] >= final_width
        or new_row["Y1"] >= final_height
        or new_row["Y2"] >= final_height
    )

    if is_out_of_bounds:
        print(f"Warning: Transformed coordinates out of bounds for file {row['HashedFileName']}")

    # Convert back to string if original was string
    for k in ["X1", "X2", "Y1", "Y2"]:
        new_row[k] = str(new_row[k])

    return new_row


def convert_measurements_csv(source_csv, output_csv, cone_params_csv=None):
    """Convert measurements CSV file with updated coordinates using cone parameters.

    Args:
        source_csv: Path to source CSV file
        output_csv: Path to output CSV file
        cone_params_csv: Path to CSV file with cone parameters
    """
    try:
        # Read the CSV file
        with open(source_csv, newline="", encoding="utf-8") as csvfile:
            reader = csv.DictReader(csvfile)
            rows = list(reader)
            fieldnames = reader.fieldnames

        # Load cone parameters if available
        cone_parameters = {}
        if cone_params_csv and Path(cone_params_csv).exists():
            cone_parameters = load_cone_parameters(cone_params_csv)
        else:
            print("Warning: No cone parameters file found. Measurements will not be transformed.")

        # Apply coordinate transformation and track skipped rows
        transformed_rows = []
        skipped_files = set()

        for row in rows:
            try:
                avi_filename = row["HashedFileName"] + ".avi"
                cone_params = cone_parameters.get(avi_filename, None)
                transformed_row = transform_measurement_coordinates_with_cone_params(
                    row, cone_params
                )
                if transformed_row is not None:
                    transformed_rows.append(transformed_row)
                else:
                    skipped_files.add(row["HashedFileName"])
            except Exception as e:
                print(f"Error processing row for file {row['HashedFileName']}: {str(e)}")
                skipped_files.add(row["HashedFileName"])

        # Save to new CSV file
        if transformed_rows:
            # Use keys from first row as fieldnames
            out_fieldnames = list(transformed_rows[0].keys())
            with open(output_csv, "w", newline="", encoding="utf-8") as csvfile:
                writer = csv.DictWriter(csvfile, fieldnames=out_fieldnames)
                writer.writeheader()
                writer.writerows(transformed_rows)
        else:
            # Write header only if no rows
            with open(output_csv, "w", newline="", encoding="utf-8") as csvfile:
                writer = csv.DictWriter(csvfile, fieldnames=fieldnames)
                writer.writeheader()

        # Print summary
        print("\nConversion Summary:")
        print(f"Total rows processed: {len(rows)}")
        print(f"Rows successfully converted: {len(transformed_rows)}")
        print(f"Rows skipped: {len(rows) - len(transformed_rows)}")
        if skipped_files:
            print("\nSkipped files:")
            for filename in sorted(skipped_files):
                print(f"  - {filename}")
        print(f"\nConverted measurements saved to {output_csv}")

    except Exception as e:
        print(f"Error processing CSV file: {str(e)}")
        raise


if __name__ == "__main__":
    args = get_args()

    # Check that cone parameters exist
    cone_params_csv = Path(args.output) / "cone_parameters.csv"
    if not cone_params_csv.exists():
        print(f"Error: Cone parameters not found at {cone_params_csv}")
        print("Please run precompute_crop.py first to generate the parameters.")
        sys.exit(1)

    # If no specific conversion is requested, convert both
    if not (args.convert_measurements or args.convert_images):
        args.convert_measurements = True
        args.convert_images = True

    # Convert images if requested
    if args.convert_images:
        source_path = Path(args.source)
        splits = load_splits(source_path)

        # Load precomputed cone parameters
        cone_parameters = load_cone_parameters(cone_params_csv)
        print(f"Loaded cone parameters for {len(cone_parameters)} files")

        files_to_process = []
        for split_files in splits.values():
            for avi_filename in split_files:
                # Strip .avi if present
                base_filename = avi_filename[:-4] if avi_filename.endswith(".avi") else avi_filename
                avi_file = find_avi_file(args.source, base_filename, batch=args.batch)
                if avi_file:
                    files_to_process.append(avi_file)
                else:
                    print(
                        f"Warning: Could not find AVI file for {base_filename} in batch "
                        f"{args.batch if args.batch else 'any'}"
                    )

        # List files that have already been processed
        files_done = []
        for _, _, filenames in os.walk(args.output):
            for filename in filenames:
                if filename.endswith(".hdf5"):
                    files_done.append(filename.replace(".hdf5", ""))

        # Filter out already processed files
        files_to_process = [f for f in files_to_process if f.stem not in files_done]

        # Limit files if max_files is specified
        if args.max_files is not None:
            files_to_process = files_to_process[: args.max_files]
            print(f"Limited to processing {args.max_files} files due to max_files parameter")

        print(f"Files left to process: {len(files_to_process)}")

        # Initialize processor with splits and cone parameters
        processor = LVHProcessor(
            path_out_h5=args.output,
            path_out=args.output_numpy,
            splits=splits,
            cone_params=cone_parameters,
        )

        print("Starting the conversion process.")

        if args.use_hyperthreading:
            with ProcessPoolExecutor() as executor:
                futures = {executor.submit(processor, file): file for file in files_to_process}
                for future in tqdm(as_completed(futures), total=len(files_to_process)):
                    try:
                        future.result()
                    except Exception as e:
                        print(f"Error processing file: {str(e)}")
        else:
            for file in tqdm(files_to_process):
                try:
                    processor(file)
                except Exception as e:
                    print(f"Error processing {file}: {str(e)}")

        print("All image conversion tasks are completed.")

    # Convert measurements if requested
    if args.convert_measurements:
        source_path = Path(args.source)
        measurements_csv = source_path / "MeasurementsList.csv"
        if measurements_csv.exists():
            output_csv = Path(args.output) / "MeasurementsList.csv"
            convert_measurements_csv(measurements_csv, output_csv, cone_params_csv)
        else:
            print("Warning: MeasurementsList.csv not found in source directory")

    print("All tasks are completed.")



"""Container for custom loss functions."""

import keras
from keras import ops


class SMSLE(keras.losses.Loss):
    """Loss function for calculating the Signed-Mean-Squared-Logarithmic-Error. This loss function
    calculates the the mean squared error on log-scaled data, and then takes the sign of the
    difference between the predicted and ground truth values into account.
    See https://doi.org/10.1109/TMI.2020.3008537 for more information.
    """

    def __init__(self, dynamic_range=60, name="smsle", **kwargs):
        super().__init__(name=name, **kwargs)
        self.dynamic_range = dynamic_range

    def call(self, y_true, y_pred):
        """
        Args:
            y_true (tensor): Ground truth values.
            y_pred (tensor): The predicted values.
        returns:
            loss (tensor): SMSLE loss value.
        """

        y_pred_max = ops.maximum(ops.max(ops.abs(y_pred)), keras.config.epsilon())
        y_true_max = ops.maximum(ops.max(ops.abs(y_true)), keras.config.epsilon())

        first_log_pos = ops.clip(
            20
            * ops.log(ops.clip(y_pred / y_pred_max, keras.config.epsilon(), 1) + 0.0)
            / ops.log(10),
            -self.dynamic_range,
            0,
        )
        secon_log_pos = ops.clip(
            20
            * ops.log(ops.clip(y_true / y_true_max, keras.config.epsilon(), 1) + 0.0)
            / ops.log(10),
            -self.dynamic_range,
            0,
        )

        first_log_neg = ops.clip(
            20
            * ops.log(ops.clip(-y_pred / y_pred_max, keras.config.epsilon(), 1) + 0.0)
            / ops.log(10),
            -self.dynamic_range,
            0,
        )
        secon_log_neg = ops.clip(
            20
            * ops.log(ops.clip(-y_true / y_true_max, keras.config.epsilon(), 1) + 0.0)
            / ops.log(10),
            -self.dynamic_range,
            0,
        )

        loss = 0.5 * ops.mean(ops.square(first_log_pos - secon_log_pos)) + 0.5 * ops.mean(
            ops.square(first_log_neg - secon_log_neg)
        )

        return loss



"""Tensorflow Ultrasound Beamforming Library.

Initialize modules for registries.
"""

import sys
from pathlib import PosixPath

import numpy as np

# Convert PosixPath objects to strings in sys.path
# this is necessary due to weird TF bug when importing
sys.path = [str(p) if isinstance(p, PosixPath) else p for p in sys.path]

import tensorflow as tf  # noqa: E402

from .dataloader import make_dataloader  # noqa: E402


def on_device_tf(func, inputs, device, return_numpy=False, **kwargs):
    """Applies a Tensorflow function to inputs on a specified device.

    Args:
        func (function): Function to apply to the input data.
        inputs (ndarray): Input array.
        device (str): Device string, e.g. ``'cuda'``, ``'gpu'``, or ``'cpu'``.
        return_numpy (bool, optional): Whether to convert output
            data back to numpy. Defaults to False.
        **kwargs: Additional keyword arguments to be passed to the ``func``.

    Returns:
        tf.Tensor or ndarray: The output data.

    Raises:
        AssertionError: If ``func`` is not a function from the tensorflow library.

    Note:
        This function converts the ``inputs`` array to a tf.Tensor and moves
        it to the specified ``device``. It then applies the ``func`` function to the inputs
        and returns the output data. If the output is a dictionary, it extracts the first value
        from the dictionary. If ``return_numpy`` is True, it converts the output data back to a
        numpy array before returning.

    Example:
        .. code-block:: python

            import tensorflow as tf


            def square(x):
                return x**2


            inputs = [1, 2, 3, 4, 5]
            device = "cuda"
            output = on_device_tf(square, inputs, device)
    """
    device = device.replace("cuda", "gpu")

    with tf.device(device):
        outputs = func(inputs, **kwargs)

    if return_numpy:
        if not isinstance(outputs, np.ndarray):
            outputs = outputs.numpy()
    return outputs



"""Utilities for models"""

import keras


class LossTrackerWrapper:
    """A wrapper for Keras Mean metrics to track multiple loss values."""

    def __init__(self, prefix):
        """
        Initialize the loss tracker wrapper.

        Args:
            prefix (str): Prefix to use for the loss name. For example "n_loss" or "i_loss".
        """
        self.prefix = prefix
        self.trackers = {}

    def update_state(self, loss_value):
        """
        Update the tracker(s) with a loss value.

        If loss_value is a dict, then for each key a separate tracker is
        created (if not already created) and updated. The tracker's name will
        be <prefix>_<key>. If loss_value is not a dict, then a default tracker
        with name <prefix> is updated.

        Args:
            loss_value: A tensor or a dictionary mapping field names to tensors.
        """
        if isinstance(loss_value, dict):
            for key, value in loss_value.items():
                tracker_name = f"{self.prefix}_{key}"
                if tracker_name not in self.trackers:
                    self.trackers[tracker_name] = keras.metrics.Mean(name=tracker_name)
                self.trackers[tracker_name].update_state(value)
        else:
            if self.prefix not in self.trackers:
                self.trackers[self.prefix] = keras.metrics.Mean(name=self.prefix)
            self.trackers[self.prefix].update_state(loss_value)

    def result(self):
        """
        Return a dictionary with the current average results.
        """
        results = {}
        for _, tracker in self.trackers.items():
            # Use the tracker's name (e.g. "n_loss_a") if available
            results[tracker.name] = tracker.result()
        return results

    def reset_state(self):
        """
        Reset all the internal trackers.
        """
        for tracker in self.trackers.values():
            tracker.reset_state()

    def __iter__(self):
        return iter(self.trackers.values())



"""Utility functions for handling local data paths.

This module provides utilities for managing local and remote data paths in ``zea`` projects.
It supports user- and machine-specific configuration via a ``users.yaml`` file, allowing
dynamic resolution of data roots for portable and reproducible workflows.

See the notebook :doc:`../notebooks/data/zea_local_data` for an extensive example of how
to set up your local data paths.


Example usage
^^^^^^^^^^^^^

.. code-block:: python

    from zea.datapaths import set_data_paths

    user = set_data_paths("users.yaml")
    print(user.data_root)

"""

import copy
import getpass
import importlib.resources
import os
import platform
import socket
import warnings
from functools import reduce
from pathlib import Path
from typing import Union

import yaml

from zea import log
from zea.config import Config
from zea.data.preset_utils import HF_PREFIX
from zea.tools.hf import HFPath
from zea.utils import strtobool

DEFAULT_DATA_ROOT = {
    "windows": "Z:/data",
    "linux": "/mnt/z/data",
    "darwin": "/mnt/z/data",
    None: "/mnt/z/data",  # for other system
}

DEFAULT_LINUX_DATA_ROOT = DEFAULT_DATA_ROOT["linux"]
DEFAULT_USERS_CONFIG_PATH = "./users.yaml"
DEFAULT_OUTPUT_PATH = "{data_root}/output"


class NoYamlFileError(Warning):
    """Raised when the users.yaml file is not found."""


class UnknownUsernameWarning(UserWarning):
    """
    Custom Warning indicating that the username was not found
    in the user.yaml file
    """


class UnknownHostnameWarning(UserWarning):
    """
    Custom Warning indicating that the hostname was not found
    for this user in the user.yaml file
    """


class UnknownLocalRemoteWarning(UserWarning):
    """
    Custom Warning indicating that the data_root corresponding to
    the local or remote key was not found
    in the user.yaml file
    """


def _create_empty_yaml(path):
    # Create empty file if it does not exist
    with open(path, "a", encoding="utf-8"):
        pass


def _fallback_to_default_data_root(system):
    if system not in DEFAULT_DATA_ROOT:
        system = None
    return DEFAULT_DATA_ROOT[system]


def _default_output_path(data_root):
    return Path(DEFAULT_OUTPUT_PATH.format(data_root=data_root))


def _verify_user_config_and_get_paths(config, system, local):
    """
    Get the user configuration and verify the paths.

    Args:
        config (dict): The configuration dictionary containing user information.
        system (str): The current operating system.
        local (bool): Flag indicating whether to use local paths or remote paths.

    Returns:
        dict: A dictionary containing the verified paths.
    """
    # Check if set os system matches with the current system
    if "system" in config:
        assert config["system"] == system, (
            f"Current OS {system} does not match user settings: {config['system']}"
        )
        config.pop("system")

    # Only keep data_root and output keys, the rest are ignored.
    unknown_keys = [x for x in config.keys() if x not in ["data_root", "output"]]
    for key in unknown_keys:
        del config[key]

    def _error_msg(key):
        return (
            f"{key} key should be either a string or a dict containing "
            "local and / or remote keys with data_root paths as values."
        )

    paths = {}
    # config will contain the data_root and optionally output paths
    for key, path in config.items():
        assert isinstance(path, (str, dict)), _error_msg(key)

        if isinstance(path, str):
            paths[key] = path
            continue

        assert set(path.keys()) <= set(["local", "remote"]), _error_msg(key)
        if local is True:
            if "local" in path:
                paths[key] = path["local"]
            else:
                warnings.warn(
                    f"Unknown local path for {key} in user config. Falling back to default.",
                    UnknownLocalRemoteWarning,
                )
                paths[key] = _fallback_to_default_data_root(system)

        elif local is False:
            if "remote" in path:
                paths[key] = path["remote"]
            else:
                warnings.warn(
                    f"Unknown remote path for {key} in user config. Falling back to default.",
                    UnknownLocalRemoteWarning,
                )
                paths[key] = _fallback_to_default_data_root(system)
        else:
            raise ValueError(
                f"Please set local to True or False or have the {key} "
                "specified as a string (without local / remote sub keys). "
                f"Current value, 'data_root': {path}."
            )

    # Set output path if not set
    if "output" not in paths:
        paths["output"] = _default_output_path(paths["data_root"])
        log.warning("No output path set, using data_root/output as output path.")

    return paths["data_root"], paths["output"]


def _verify_paths(data_path):
    """Verify that the paths exist and are directories."""
    for key in ["data_root", "output"]:
        path = data_path[key]
        if not Path(path).is_dir():
            log.warning(
                f"{key} path `{path}` does not exist, please update your "
                f"{log.yellow('users.yaml')} file."
            )


def _load_users_yaml(user_config, local, username, hostname):
    config_path = Path(user_config)

    # If there is no users.yaml file yet, create one.
    if not config_path.is_file():
        warnings.warn(
            f"No {user_config} file found, creating a new one. "
            "Consider running `python -m zea.datapaths` to setup your paths. ",
            NoYamlFileError,
        )

        _create_empty_yaml(config_path)

        try:
            create_new_user(local=local)
        except Exception:
            log.warning(
                f"Could not create user profile for {username} on {hostname}, using default."
            )

    # Load YAML file with user info
    with open(config_path, "r", encoding="utf-8") as file:
        config = yaml.safe_load(file)

    if config is None:
        config = {}

    if not isinstance(config, dict):
        # Raise error if config is not a dictionary, for example if its empty.
        # Lets not overwrite the users config file in this case.
        raise ValueError(
            f"""YAML file should contain a dictionary, but found {type(config)}".
            Please check your users.yaml file for corruptions. In case you want to create a
            new users.yaml file, please delete the current one."""
        )
    return config


def set_data_paths(
    user_config: Union[str, dict] = None, local: bool = True, verify: bool = True
) -> dict:
    """Get data paths (absolute paths to location of data).

    Args:
        user_config (str or dict, optional): Path to a YAML file with user info.
            If None, uses ``./users.yaml`` as the default file. Can also be a dictionary
            structured as shown below.
        local (bool, optional): Use local dataset or get from NAS.
        verify (bool, optional): Verify that the paths exist and are directories.
            Default is True.

    Example YAML structure::

        data_root: ...
        output: ...

    You can also specify different ``data_root`` for different users and machines::

        my_username:
          my_hostname:
            system: windows
            data_root: ...
            output: ...
          other_hostname:
            system: linux
            data_root:
              local: ...
              remote: ...
          # If both my_hostname and other_hostname are not matching, fallback to:
          system: linux
          data_root: ...

        other_username:
          data_root: ...

    These will take precedence over the ``data_root`` that is userless and machineless.

    Returns:
        dict: Absolute paths to location of data. Stores the following parameters:
            ``data_root``, ``repo_root``, ``output``, ``system``, ``username``, ``hostname``

    """
    username = getpass.getuser()
    system = platform.system().lower()
    hostname = socket.gethostname()
    repo_root = importlib.resources.files("zea")  # ultrasound-toolbox/zea

    # If user_config is None, use the default users.yaml file
    if isinstance(user_config, type(None)):
        user_config = DEFAULT_USERS_CONFIG_PATH

    # If user_config is a dictionary, use it as the config
    if isinstance(user_config, dict):
        config = copy.deepcopy(user_config)
    # If user_config is a string, load the yaml file
    elif isinstance(user_config, str):
        config = _load_users_yaml(user_config, local, username, hostname)
    else:
        raise ValueError("user_config should be a string or dictionary.")

    # Check if username is in the config, if so, select that part of the config
    if username in config:
        config = config[username]

    # Check if hostname is in the config, if so, select that part of the config
    if hostname in config:
        config = config[hostname]

    # Ensure that the remaining config contains a `data_root` key
    if "data_root" not in config:
        warnings.warn(
            (
                f"Cannot find data_root for username={username} "
                f"and hostname={hostname} in user file. Also no default data_root found. "
                f"Falling back to default path for {system}: {DEFAULT_DATA_ROOT[system]}. "
                f"Please update the `{user_config}` with your data-path settings."
            ),
            UnknownUsernameWarning,
        )
        data_root = _fallback_to_default_data_root(system)
        output = _default_output_path(data_root)
    else:
        data_root, output = _verify_user_config_and_get_paths(config, system, local)

    data_path = {
        "data_root": Path(data_root),
        "repo_root": repo_root,
        "output": Path(output),
        "system": system,
        "username": username,
        "hostname": hostname,
    }

    if verify:
        _verify_paths(data_path)

    return Config(data_path)


## Helper functions for handling user input


def _build_user_profile_string(data_paths, local: bool = None):
    """Builds a string that can be written to users.yaml to create a new user profile."""
    tab = "    "  # 4 spaces required in yaml
    base_string = (
        f"'{data_paths['username']}':\n"
        + f"  {data_paths['hostname']}:\n"
        + f"    system: {data_paths['system']}\n"
    )
    if local is None:
        return base_string + f"{tab}data_root: {data_paths['data_root']}"
    elif local is False:
        return base_string + (f"{tab}data_root:\n" + f"{tab}{tab}remote: {data_paths['data_root']}")
    elif local is True:
        return base_string + (f"{tab}data_root:\n" + f"{tab}{tab}local: {data_paths['data_root']}")
    else:
        raise ValueError("local should set to a boolean or None.")


def _to_write_user_profile_to_file(user_profile_string, user_config_path=DEFAULT_USERS_CONFIG_PATH):
    with open(user_config_path, "a", encoding="utf-8") as file:
        file.write("\n\n" + user_profile_string + "\n")
    print(f"\n✅ Your user profile was successfully added to `{user_config_path}`.\n")


def _pretty_print_data_paths(data_paths):
    for key, value in data_paths.items():
        print(f"\t{key}: {log.yellow(value)}")
    print()


def _prompt_user_for_data_root():
    data_root_input = input(
        "\nℹ️  Please enter the path to your data directory, "
        "or press Enter to use the default Linux path "
        f"`{DEFAULT_LINUX_DATA_ROOT}`: "
    )
    return DEFAULT_LINUX_DATA_ROOT if data_root_input == "" else data_root_input


def _acquire_and_validate_data_root():
    data_root_input = _prompt_user_for_data_root()
    while not os.path.isdir(data_root_input):
        print("\n The path you entered does not point to a directory, please try again.")
        data_root_input = _prompt_user_for_data_root()
    return data_root_input


def _warning_type_was_thrown(warning_type, list_of_warnings):
    """Returns True iff list_of_warnings contains a warning of type warning_type"""
    if not list_of_warnings:
        return False
    return reduce(
        lambda acc, w: acc and isinstance(w.message, warning_type),
        list_of_warnings,
        True,
    )


def _to_read_yaml_file(path_str):
    path = Path(path_str)
    if not path.is_file():
        raise ValueError("YAML file path provided does not lead to a file.")

    with open(path, "r", encoding="utf-8") as file:
        config = yaml.safe_load(file)
        return config


def _to_write_yaml_file(data, path_str):
    path = Path(path_str)
    if _check_for_comments_yaml_file(path_str):
        log.warning(
            f"YAML file {path_str} contains comments. "
            "These will be removed if you write to the file."
        )
        input("Press Enter to continue or Ctrl+C to cancel.")

    if not path.is_file():
        raise ValueError("YAML file path provided does not lead to a file.")

    with open(path, "w", encoding="utf-8") as file:
        yaml.dump(data, file, default_flow_style=False, sort_keys=False)


def _try(fn, args):
    try:
        return fn(**args)
    except Exception as e:
        print(f"Encountered an error in {fn.__name__}")
        print(e)


def _check_for_comments_yaml_file(path_str):
    """Returns True iff the YAML file at path_str contains comments."""
    path = Path(path_str)
    if not path.is_file():
        raise ValueError("YAML file path provided does not lead to a file.")

    with open(path, "r", encoding="utf-8") as file:
        lines = file.readlines()
        # just look for # anywhere
        return any("#" in line for line in lines)


def create_new_user(user_config_path: str = None, local: bool = None):
    """Creates a new user profile in `users.yaml` if one does not already exist.

    Args:
        user_config (str): Path that points to yaml file with user info.
            Defaults to None. In that case `./users.yaml` is taken
        local (bool): Use local dataset or get from remote (NAS).
            Per machine, the data_root can be set to a local or remote path.
            Each user can also have a different data_root for each machine.
            Default is None, which means that the data_root is shared for either
            local or remote (i.e. this parameter is ignored), see doc set_data_paths().
    """
    # Create empty file if it does not exist
    _create_empty_yaml(user_config_path)

    with warnings.catch_warnings(record=True) as list_of_warnings:
        data_paths = set_data_paths(user_config=user_config_path, local=local)
        if user_config_path is None:
            user_config_path = DEFAULT_USERS_CONFIG_PATH
        assert isinstance(user_config_path, str), "user_config_path should be a string."

        # Display any warnings that were thrown during set_data_paths
        if list_of_warnings:
            for w in list_of_warnings:
                print(f"🚨 {w.message}")
        else:
            log.info("Data paths set successfully.")
            log.info("Here's a summary of your data paths:")
            _pretty_print_data_paths(data_paths)

        # If there was no profile found in users.yaml for the current user,
        # give them the option to create a user profile automatically
        user_warning_was_thrown = _warning_type_was_thrown(UnknownUsernameWarning, list_of_warnings)
        hostname_warning_was_thrown = _warning_type_was_thrown(
            UnknownHostnameWarning, list_of_warnings
        )
        local_remote_warning_was_thrown = _warning_type_was_thrown(
            UnknownLocalRemoteWarning, list_of_warnings
        )
        no_yaml_file_error_was_thrown = _warning_type_was_thrown(NoYamlFileError, list_of_warnings)

        if user_warning_was_thrown or no_yaml_file_error_was_thrown:
            print("ℹ️  Follow the instructions below to create your user profile.")
            data_root = _acquire_and_validate_data_root()
            data_paths["data_root"] = data_root
            user_profile_string = _build_user_profile_string(data_paths, local=local)
            user_response = input(
                "\n"
                + user_profile_string
                + "\n"
                + "\nℹ️  Would you like to automatically create your user"
                + "profile with the details above? [y]: "
            )
            if user_response == "" or strtobool(user_response):
                _try(
                    _to_write_user_profile_to_file,
                    {
                        "user_profile_string": user_profile_string,
                        "user_config_path": user_config_path,
                    },
                )
        elif hostname_warning_was_thrown:
            print(
                f"ℹ️ The hostname '{data_paths['hostname']}' was "
                f"not found for username '{data_paths['username']}'.\n"
            )
            print(
                "ℹ️ Follow the instructions below to create a new "
                f"entry for hostname: '{data_paths['hostname']}:"
            )
            data_root = _acquire_and_validate_data_root()
            data_paths["data_root"] = data_root
            users_yaml_dict = _try(_to_read_yaml_file, {"path_str": user_config_path})
            users_yaml_dict[data_paths["username"]][data_paths["hostname"]] = {
                "system": data_paths["system"],
                "data_root": data_root,
            }
            user_response = input(
                "\n"
                + yaml.dump(users_yaml_dict[data_paths["username"]])
                + "\nℹ️ Would you like to update your user profile "
                + "with the user info above? [y]: "
            )
            if user_response == "" or strtobool(user_response):
                _try(
                    _to_write_yaml_file,
                    {"data": users_yaml_dict, "path_str": user_config_path},
                )
                log.success("Profile updated successfully.")
        elif local_remote_warning_was_thrown:
            local_remote_str = "local" if local else "remote"
            print(
                f"ℹ️ The data_root for '{data_paths['username']}' was "
                f"not found for location: {local_remote_str}.\n"
            )
            print(
                "ℹ️ Follow the instructions below to create a new entry for "
                f"data_root for location: {local_remote_str}:"
            )
            data_root = _acquire_and_validate_data_root()
            data_paths["data_root"] = data_root
            users_yaml_dict = _try(_to_read_yaml_file, {"path_str": user_config_path})
            ## now update the data_root for the user and hostname in the yaml file
            ## use local or remote subkey depending on the local parameter
            users_yaml_dict[data_paths["username"]][data_paths["hostname"]]["data_root"].update(
                {local_remote_str: data_root}
            )
            user_response = input(
                "\n"
                + yaml.dump(users_yaml_dict[data_paths["username"]])
                + "\nℹ️ Would you like to update your user profile "
                + "with the user info above? [y]: "
            )
            if user_response == "" or strtobool(user_response):
                _try(
                    _to_write_yaml_file,
                    {"data": users_yaml_dict, "path_str": user_config_path},
                )
                log.success("Profile updated successfully.")

    return data_paths


def format_data_path(path: str, user: Config = None) -> Path:
    """If the path is not absolute, prepend the data_root to it."""
    if Path(path).is_absolute():
        return Path(path)
    elif str(path).startswith(HF_PREFIX):
        return HFPath(path)
    else:
        assert user is not None, (
            "The dataset folder is relative, but no user is provided. "
            "Please provide a user to load the dataset relative to "
            "the user's data_root."
        )
    return Path(user.data_root) / path


if __name__ == "__main__":
    create_new_user("users.yaml", local=None)



"""UNet models and architectures"""

import keras
from keras import layers

from zea import log
from zea.internal.registry import model_registry
from zea.models.base import BaseModel
from zea.models.layers import DownBlock, ResidualBlock, UpBlock, sinusoidal_embedding
from zea.models.preset_utils import register_presets
from zea.models.presets import unet_presets


@model_registry(name="unet")
class UNet(BaseModel):
    """UNet model"""

    def __init__(
        self,
        input_shape,
        widths,
        block_depth,
        input_range,
        name="unet",
        **kwargs,
    ):
        """Initializes a UNet model"""

        super().__init__(name=name, **kwargs)

        self.input_shape = input_shape
        self.input_range = input_range
        self.widths = widths
        self.block_depth = block_depth

        self.network = get_unetwork(self.input_shape, self.widths, self.block_depth)

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "input_shape": self.input_shape,
                "input_range": self.input_range,
                "widths": self.widths,
                "block_depth": self.block_depth,
            }
        )
        return config

    def call(self, *args, **kwargs):
        return self.network(*args, **kwargs)


def get_unetwork(
    image_shape,
    widths,
    block_depth,
):
    """Get a basic UNet architecture

    Args:
        image_shape: tuple, (height, width, channels)
        widths: list, number of filters in each layer
        block_depth: int, number of residual blocks in each down/up block

    Returns:
        keras.Model
    """
    assert len(image_shape) == 3, "image_shape must be a tuple of (height, width, channels)"

    image_height, image_width, n_channels = image_shape
    noisy_images = keras.Input(shape=(image_height, image_width, n_channels))

    x = layers.Conv2D(widths[0], kernel_size=1)(noisy_images)

    skips = []
    for width in widths[:-1]:
        x = DownBlock(width, block_depth)([x, skips])

    for _ in range(block_depth):
        x = ResidualBlock(widths[-1])(x)

    for width in reversed(widths[:-1]):
        x = UpBlock(width, block_depth)([x, skips])

    x = layers.Conv2D(n_channels, kernel_size=1, kernel_initializer="zeros")(x)

    return keras.Model(noisy_images, x, name="residual_unet")


@model_registry(name="unet_time_conditional")
class UNetTimeConditional(BaseModel):
    """UNet model with time-conditional sinusoidal embedding"""

    def __init__(
        self,
        image_shape,
        widths,
        block_depth,
        image_range,
        embedding_min_frequency=1.0,
        embedding_max_frequency=1000.0,
        embedding_dims=32,
        name="unet_time_conditional",
        **kwargs,
    ):
        super().__init__(name=name, **kwargs)
        self.image_shape = image_shape
        self.image_range = image_range
        self.widths = widths
        self.block_depth = block_depth
        self.embedding_min_frequency = embedding_min_frequency
        self.embedding_max_frequency = embedding_max_frequency
        self.embedding_dims = embedding_dims
        self.network = get_time_conditional_unetwork(
            self.image_shape,
            self.widths,
            self.block_depth,
            self.embedding_min_frequency,
            self.embedding_max_frequency,
            self.embedding_dims,
        )

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "image_shape": self.image_shape,
                "image_range": self.image_range,
                "widths": self.widths,
                "block_depth": self.block_depth,
                "embedding_min_frequency": self.embedding_min_frequency,
                "embedding_max_frequency": self.embedding_max_frequency,
                "embedding_dims": self.embedding_dims,
            }
        )
        return config

    def call(self, *args, **kwargs):
        return self.network(*args, **kwargs)


def get_time_conditional_unetwork(
    image_shape,
    widths=None,
    block_depth=None,
    embedding_min_frequency=1.0,
    embedding_max_frequency=1000.0,
    embedding_dims=32,
):
    """Get a basic UNet architecture with time-conditional sinusoidal embeddings

    Used in Diffusion Models.

    Args:
        image_shape: tuple, (height, width, channels)
        widths: list, number of filters in each layer
        block_depth: int, number of residual blocks in each down/up block
        embedding_min_frequency: float, minimum frequency for sinusoidal embeddings
        embedding_max_frequency: float, maximum frequency for sinusoidal embeddings
        embedding_dims: int, number of dimensions for sinusoidal embeddings

    Returns:
        keras.Model
    """
    assert len(image_shape) == 3, "image_shape must be a tuple of (height, width, channels)"

    if widths is None:
        log.warning("No widths provided, using default widths [32, 64, 96, 128]")
        widths = [32, 64, 96, 128]
    if block_depth is None:
        log.warning("No block_depth provided, using default block_depth 2")
        block_depth = 2

    image_height, image_width, n_channels = image_shape
    noisy_images = keras.Input(shape=(image_height, image_width, n_channels))
    noise_variances = keras.Input(shape=(1, 1, 1))

    @keras.saving.register_keras_serializable()
    def _sinusoidal_embedding(x):
        return sinusoidal_embedding(
            x, embedding_min_frequency, embedding_max_frequency, embedding_dims
        )

    e = layers.Lambda(_sinusoidal_embedding, output_shape=(1, 1, 32))(noise_variances)
    e = layers.UpSampling2D(size=(image_height, image_width), interpolation="nearest")(e)

    x = layers.Conv2D(widths[0], kernel_size=1)(noisy_images)
    x = layers.Concatenate()([x, e])

    skips = []
    for width in widths[:-1]:
        x = DownBlock(width, block_depth)([x, skips])

    for _ in range(block_depth):
        x = ResidualBlock(widths[-1])(x)

    for width in reversed(widths[:-1]):
        x = UpBlock(width, block_depth)([x, skips])

    x = layers.Conv2D(n_channels, kernel_size=1, kernel_initializer="zeros")(x)

    return keras.Model([noisy_images, noise_variances], x, name="residual_unet")


register_presets(unet_presets, UNet)
register_presets(unet_presets, UNetTimeConditional)



"""Model and constructor for unfolded convolutional ISTA"""

import keras
from keras import Input, ops
from keras.layers import Add, Conv2D, UpSampling2D
from tensorflow.python.keras.layers import Layer

from zea.backend.tensorflow.layers.utils import get_activation


def UnfoldingModel(
    input_dim,
    activation=None,
    folds=5,
    upsampling=1,
    P0_dim=None,
    P1_dim=None,
):
    """Unfolding Iterative Shrinking and Thresholding Model

    Args:
        input_dim (tuple): input dimensions
        activation (str, optional): final activation function.
            Defaults to None.
        folds (int, optional): number of folds. Defaults to 5.
        upsampling (int, optional): upsampling of output. Defaults to 1.
        P0_dim (list, optional): dimensions of conv kernels.
            Defaults to [1, 5].
        P1_dim (list, optional): dimensions of conv kernels.
            Defaults to [1, 5].

    Returns:
        tf model: unfolded LISTA model
    """
    if P0_dim is None:
        P0_dim = [1, 5]
    if P1_dim is None:
        P1_dim = [1, 5]

    inp = Input(shape=(input_dim[0], input_dim[1], 1))

    input_up = UpSampling2D(
        size=(upsampling, upsampling), interpolation="nearest", name="Upsample"
    )(inp)

    xP0 = Conv2D(
        P0_dim[0],
        (P0_dim[1], P0_dim[1]),
        activation=None,
        padding="same",
        strides=[1, 1],
        name="x_P0",
    )(input_up)

    x = xP0
    for k in range(1, folds):
        x_thresh = Prox(name=f"x_thresh_{k}")(x)
        x_thresh_P1 = Conv2D(
            P1_dim[0],
            (P1_dim[1], P1_dim[1]),
            activation=None,
            padding="same",
            strides=[1, 1],
            name=f"x_thresh_P1_{k}",
        )(x_thresh)

        xP0 = Conv2D(
            P0_dim[0],
            (P0_dim[1], P0_dim[1]),
            activation=None,
            padding="same",
            strides=[1, 1],
            name=f"x_P0_{k}",
        )(input_up)
        x = Add(name=f"x_{k}")([x_thresh_P1, xP0])

    x_thresh = Prox(name="S_out")(x)

    out = Conv2D(1, (1, 1), activation=None, padding="same", strides=[1, 1], name="SP1_out")(
        x_thresh
    )
    out = get_activation(activation)(out)

    return keras.Model(inputs=inp, outputs=out)


class Prox(Layer):
    """Proximal layer"""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.alpha = None

    def build(self, input_shape):
        self.alpha = self.add_weight(shape=(1, 1), initializer="random_normal", trainable=True)
        super().build(input_shape)

    def call(self, inputs, *args, **kwargs):
        """Apply the proximal operator.

        Args:
            inputs (Tensor): Input tensor.
            *args: Additional positional arguments.
            **kwargs: Additional keyword arguments.

        Returns:
            Tensor: Output after applying the proximal operator.
        """
        return ops.sign(inputs) * ops.relu(ops.abs(inputs) - ops.softplus(self.alpha))

    def compute_output_shape(self, input_shape):
        return input_shape



"""Check functions for data types and shapes."""

import importlib.util

from zea.internal.core import DataTypes, ModTypes
from zea.internal.registry import checks_registry

_DATA_TYPES = [member.value for member in DataTypes]
_MOD_TYPES = [member.value for member in ModTypes]

_BACKENDS = [None, "torch", "tensorflow", "keras", "jax", "numpy"]

_ML_LIB_AVAILABLE = False
for lib in _BACKENDS:
    if importlib.util.find_spec(str(lib)):
        if lib == "torch":
            _ML_LIB_AVAILABLE = True
        if lib == "tensorflow":
            _ML_LIB_AVAILABLE = True

_REQUIRED_SCAN_KEYS = [
    "n_ax",
    "n_el",
    "n_tx",
    "probe_geometry",
    "sampling_frequency",
    "center_frequency",
    "t0_delays",
    "n_frames",
]

_IMAGE_DATA_TYPES = ["image", "image_sc", "envelope_data", "beamformed_data"]

_NON_IMAGE_DATA_TYPES = ["raw_data", "aligned_data"]


def get_check(data_type):
    """Get check function for data type.

    Args:
        data_type (str): data type to get check function for

    Raises:
        ValueError: if data type is not valid

    Returns:
        function: check function for data type
    """
    if data_type not in _DATA_TYPES:
        raise ValueError(f"Data type {data_type} not valid. Must be one of {_DATA_TYPES}")
    return checks_registry[data_type]


@checks_registry("raw_data")
def _check_raw_data(data=None, shape=None, with_batch_dim=None):
    """Check raw data shape.

    If data is provided, shape is derived from data.
    If shape is provided, data is ignored. Only supply one of data or shape.

    Args:
        data (np.ndarray, optional): raw data. Defaults to None.
            either data or shape must be provided.
        shape (tuple, optional): shape of the data. Defaults to None.
            either data or shape must be provided.
        with_batch_dim (bool, optional): whether data has frame dimension at the start.
            Setting this to True requires the data to have 5 dimensions. Defaults to
            False.

    Raises:
        AssertionError: if data does not have expected shape
        AssertionError: if data does not have expected number of channels
    """
    assert data is not None or shape is not None, "Either data or shape must be provided."
    if data is not None:
        shape = data.shape

    if with_batch_dim is None:
        with_batch_dim = len(shape) == 5

    if not with_batch_dim:
        assert len(shape) == 4, (
            f"raw data must be 4D, with expected shape [n_tx, n_ax, n_el, n_ch], got {shape}"
        )
    else:
        assert len(shape) == 5, (
            f"raw data must be 5D, with expected shape [n_fr, n_tx, n_ax, n_el, n_ch], got {shape}"
        )
    assert shape[-1] in [1, 2], (
        "raw data must have 1 or 2 channels, for RF or IQ data respectively, "
        f"got {shape[-1]} channels"
    )


@checks_registry("aligned_data")
def _check_aligned_data(data=None, shape=None, with_batch_dim=None):
    """Check aligned data shape.

    If data is provided, shape is derived from data.
    If shape is provided, data is ignored. Only supply one of data or shape.

    Args:
        data (np.ndarray, optional): aligned data. Defaults to None.
            either data or shape must be provided.
        shape (tuple, optional): shape of the data. Defaults to None.
            either data or shape must be provided.
        with_batch_dim (bool, optional): whether data has frame dimension at the start.
            Setting this to True requires the data to have 5 dimensions. Defaults to
            False.

    Raises:
        AssertionError: if data does not have expected shape
        AssertionError: if data does not have expected number of channels
    """
    assert data is not None or shape is not None, "Either data or shape must be provided."
    if data is not None:
        shape = data.shape

    if with_batch_dim is None:
        with_batch_dim = len(shape) == 5

    if not with_batch_dim:
        assert len(shape) == 4, (
            f"aligned data must be 4D, with expected shape [n_tx, n_ax, n_el, n_ch], got {shape}"
        )
    else:
        assert len(shape) == 5, (
            "aligned data must be 5D, with expected shape [n_fr, n_tx, n_ax, n_el, n_ch], "
            f"got {shape}"
        )
    assert shape[-1] in [1, 2], (
        "raw data must have 1 or 2 channels, for RF or IQ data respectively, "
        f"got {shape[-1]} channels"
    )


@checks_registry("beamformed_data")
def _check_beamformed_data(data=None, shape=None, with_batch_dim=None):
    """Check beamformed data shape.

    If data is provided, shape is derived from data.
    If shape is provided, data is ignored. Only supply one of data or shape.

    Args:
        data (np.ndarray, optional): beamformed data. Defaults to None.
            either data or shape must be provided.
        shape (tuple, optional): shape of the data. Defaults to None.
            either data or shape must be provided.
        with_batch_dim (bool, optional): whether data has frame dimension at the start.
            Setting this to True requires the data to have 4 dimensions. Defaults to
            False.

    Raises:
        AssertionError: if data does not have expected shape
        AssertionError: if data does not have expected number of channels
    """
    assert data is not None or shape is not None, "Either data or shape must be provided."
    if data is not None:
        shape = data.shape

    if with_batch_dim is None:
        with_batch_dim = len(shape) == 4

    if not with_batch_dim:
        assert len(shape) == 3, (
            f"beamformed data must be 3D, with expected shape [Ny, Nx, n_ch], got {shape}"
        )
    else:
        assert len(shape) == 4, (
            f"beamformed data must be 4D, with expected shape [n_fr, Ny, Nx, n_ch], got {shape}"
        )
    assert shape[-1] in [1, 2], (
        "beamformed data must have 1 or 2 channels, for RF or IQ data respectively, "
        f"got {shape[-1]} channels"
    )


@checks_registry("envelope_data")
def _check_envelope_data(data=None, shape=None, with_batch_dim=None):
    """Check envelope data shape.

    If data is provided, shape is derived from data.
    If shape is provided, data is ignored. Only supply one of data or shape.

    Args:
        data (np.ndarray, optional): envelope data. Defaults to None.
            either data or shape must be provided.
        shape (tuple, optional): shape of the data. Defaults to None.
            either data or shape must be provided.
        with_batch_dim (bool, optional): whether data has frame dimension at the start.
            Setting this to True requires the data to have 4 dimensions. Defaults to
            False.

    Raises:
        AssertionError: if data does not have expected shape
    """
    assert data is not None or shape is not None, "Either data or shape must be provided."
    if data is not None:
        shape = data.shape

    if with_batch_dim is None:
        with_batch_dim = len(shape) == 3

    if not with_batch_dim:
        assert len(shape) == 2, (
            f"envelope data must be 2D, with expected shape [Ny, Nx], got {shape}"
        )
    else:
        assert len(shape) == 3, (
            f"envelope data must be 3D, with expected shape [n_fr, Ny, Nx], got {shape}"
        )


@checks_registry("image")
def _check_image(data=None, shape=None, with_batch_dim=None):
    """Check image data shape.

    If data is provided, shape is derived from data.
    If shape is provided, data is ignored. Only supply one of data or shape.

    Args:
        data (np.ndarray, optional): image data. Defaults to None.
            either data or shape must be provided.
        shape (tuple, optional): shape of the data. Defaults to None.
            either data or shape must be provided.
        with_batch_dim (bool, optional): whether data has frame dimension at the start.
            Setting this to True requires the data to have 4 dimensions. Defaults to
            False.

    Raises:
        AssertionError: if data does not have expected shape.
    """
    assert data is not None or shape is not None, "Either data or shape must be provided."
    if data is not None:
        shape = data.shape

    if with_batch_dim is None:
        with_batch_dim = len(shape) == 3

    if not with_batch_dim:
        assert len(shape) == 2, f"image data must be 2D, with expected shape [Ny, Nx], got {shape}"
    else:
        assert len(shape) == 3, (
            f"image data must be 3D, with expected shape [n_fr, Ny, Nx], got {shape}"
        )


@checks_registry("image_sc")
def _check_image_sc(data=None, shape=None, with_batch_dim=None):
    """Check image data shape.

    If data is provided, shape is derived from data.
    If shape is provided, data is ignored. Only supply one of data or shape.

    Args:
        data (np.ndarray, optional): scan-converted data. Defaults to None.
            either data or shape must be provided.
        shape (tuple, optional): shape of the data. Defaults to None.
            either data or shape must be provided.
        with_batch_dim (bool, optional): whether data has frame dimension at the start.
            Setting this to True requires the data to have 4 dimensions. Defaults to
            False.

    Raises:
        AssertionError: if data does not have expected shape.
    """
    assert data is not None or shape is not None, "Either data or shape must be provided."
    if data is not None:
        shape = data.shape

    if with_batch_dim is None:
        with_batch_dim = len(shape) == 3

    if not with_batch_dim:
        assert len(shape) == 2, f"image data must be 2D, with expected shape [Ny, Nx], got {shape}"
    else:
        assert len(shape) == 3, (
            f"image data must be 3D, with expected shape [n_fr, Ny, Nx], got {shape}"
        )


def _assert_keys_and_axes(keys, axes):
    """Quick check to ensure that the keys and axes are lists of the same length,
    and that the keys are strings and the axes are integers."""

    if not isinstance(keys, list):
        keys = [keys]
    if not isinstance(axes, list):
        axes = [axes]
    if len(keys) != len(axes):
        raise ValueError("The number of keys and axes must match.")

    # assert that all keys are strings
    for key in keys:
        assert isinstance(key, str), "All keys must be strings."

    # assert that all axes are integers
    for axis in axes:
        assert isinstance(axis, int), "All axes must be integers."

    return keys, axes



"""Diffusion models"""

import abc
from typing import Literal

import keras
from keras import ops

from zea.backend import _import_tf
from zea.backend.autograd import AutoGrad
from zea.internal.core import Object
from zea.internal.operators import Operator
from zea.internal.registry import diffusion_guidance_registry, model_registry, operator_registry
from zea.models.dense import get_time_conditional_dense_network
from zea.models.generative import DeepGenerativeModel
from zea.models.preset_utils import register_presets
from zea.models.presets import diffusion_model_presets
from zea.models.unet import get_time_conditional_unetwork
from zea.models.utils import LossTrackerWrapper
from zea.tensor_ops import L2, fori_loop, split_seed
from zea.utils import fn_requires_argument

tf = _import_tf()


@model_registry(name="diffusion")
class DiffusionModel(DeepGenerativeModel):
    """Implementation of a diffusion generative model.
    Heavily inspired from https://keras.io/examples/generative/ddim/
    """

    def __init__(
        self,
        input_shape,
        input_range=(0, 1),
        min_signal_rate=0.02,
        max_signal_rate=0.95,
        network_name="unet_time_conditional",
        network_kwargs=None,
        name="diffusion_model",
        guidance="dps",
        operator="inpainting",
        **kwargs,
    ):
        """Initialize a diffusion model.

        Args:
            input_shape: Shape of the input data. Typically of the form
                `(height, width, channels)` for images.
            widths: List of filter widths for the UNet.
            block_depth: Number of residual blocks in each UNet block.
            timesteps: Number of diffusion timesteps.
            beta_start: Initial noise schedule value.
            beta_end: Final noise schedule value.
            name: Name of the model.
            guidance: Guidance method to use. Can be a string, or dict with
                "name" and "params" keys. Additionally, can be a `DiffusionGuidance` object.
            operator: Operator to use. Can be a string, or dict with
                "name" and "params" keys. Additionally, can be a `Operator` object.

            **kwargs: Additional arguments.
        """
        super().__init__(name=name, **kwargs)

        self.input_shape = input_shape
        self.input_range = input_range
        self.min_signal_rate = min_signal_rate
        self.max_signal_rate = max_signal_rate
        self.network_name = network_name
        self.network_kwargs = network_kwargs or {}

        # reverse diffusion (i.e. sampling) goes from max_t to min_t
        self.min_t = 0.0
        self.max_t = 1.0

        if network_name == "unet_time_conditional":
            self.network = get_time_conditional_unetwork(
                image_shape=self.input_shape,
                **self.network_kwargs,
            )
        elif network_name == "dense_time_conditional":
            assert len(input_shape) == 1, "Dense network only supports 1D input"
            self.network = get_time_conditional_dense_network(
                input_dim=self.input_shape[0],
                **self.network_kwargs,
            )
        else:
            raise ValueError("Invalid network name provided.")

        # Also initialize the exponential moving average network
        self.ema_network = keras.models.clone_model(self.network)
        self.ema_network.trainable = False

        self.image_loss_tracker = LossTrackerWrapper("i_loss")
        self.noise_loss_tracker = LossTrackerWrapper("n_loss")

        # for storing intermediate results (i.e. diffusion trajectory)
        self.track_progress_interval = 1
        self.track_progress = []

        # for guidance / conditional sampling
        self.guidance_fn = None
        self.operator = None
        self._init_operator_and_guidance(operator, guidance)

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "input_shape": self.input_shape,
                "input_range": self.input_range,
                "min_signal_rate": self.min_signal_rate,
                "max_signal_rate": self.max_signal_rate,
                "network_name": self.network_name,
                "network_kwargs": self.network_kwargs,
            }
        )
        return config

    def _init_operator_and_guidance(self, operator, guidance):
        if operator is not None:
            if isinstance(operator, str):
                operator_class = operator_registry[operator]
                self.operator = operator_class()
            elif isinstance(operator, Operator):
                self.operator = operator
            elif isinstance(operator, dict):
                operator_class = operator_registry[operator["name"]]
                if "params" not in operator:
                    operator["params"] = {}
                if (
                    fn_requires_argument(operator_class.__init__, "image_range")
                    and "image_range" not in operator["params"]
                ):
                    operator["params"]["image_range"] = self.input_range
                self.operator = operator_class(**operator["params"])
            else:
                raise ValueError(
                    f"Invalid operator provided, must be a string, dict or "
                    f"Operator object, got {operator}"
                )

        if guidance is not None:
            assert operator is not None, "Operator must be provided for guidance"
            if isinstance(guidance, str):
                guidance_class = diffusion_guidance_registry[guidance]
                self.guidance_fn = guidance_class(
                    diffusion_model=self,
                    operator=self.operator,
                )
            elif isinstance(guidance, DiffusionGuidance):
                self.guidance_fn = guidance
            elif isinstance(guidance, dict):
                guidance_class = diffusion_guidance_registry[guidance["name"]]
                self.guidance_fn = guidance_class(
                    diffusion_model=self, operator=self.operator, **guidance["params"]
                )
            else:
                raise ValueError(
                    f"Invalid guidance provided, must be a string, dict or "
                    f"DiffusionGuidance object, got {guidance}"
                )

    def call(self, inputs, training=False, **kwargs):
        """
        Calls the score network.

        Will use the exponential moving average network if training is False,
        otherwise the regular network."""
        if training:
            network = self.network
        else:
            network = self.ema_network
        return network(inputs, training=training, **kwargs)

    def sample(self, n_samples=1, n_steps=20, seed=None, **kwargs):
        """Sample from the model.

        Args:
            n_samples: Number of samples to generate.
            n_steps: Number of diffusion steps.
            seed: Random seed generator.
            **kwargs: Additional arguments.

        Returns:
            Generated samples.
        """
        seed, seed1 = split_seed(seed, 2)

        # Generate random noise
        noise = keras.random.normal(
            shape=(n_samples, *self.input_shape),
            seed=seed1,
        )
        # Reverse diffusion process
        return self.reverse_diffusion(
            initial_noise=noise, diffusion_steps=n_steps, seed=seed, **kwargs
        )

    def posterior_sample(
        self,
        measurements,
        n_samples=1,
        n_steps=20,
        initial_step=0,
        initial_samples=None,
        seed=None,
        **kwargs,
    ):
        """Sample from the posterior distribution given measurements.

        Args:
            measurements: Input measurements. Typically of shape
                `(batch_size, *input_shape)`.
            n_samples: Number of posterior samples to generate.
                Will generate `n_samples` samples for each measurement
                in the `measurements` batch.
            n_steps: Number of diffusion steps.
            initial_step: Initial step to start from. Can warm start the
                diffusion process with a partially noised image, thereby
                skipping part of the diffusion process. Initial step
                closer to n_steps, will result in a shorter diffusion process
                (i.e. less noise added to the initial image). A value of 0
                means that the diffusion process starts from pure noise.
            initial_samples: Optional initial samples to start from.
                If provided, these samples will be used as the starting point
                for the diffusion process. Only used if `initial_step` is
                greater than 0. Must be of shape `(batch_size, n_samples, *input_shape)`.
            seed: Random seed generator.
            **kwargs: Additional arguments.

        Returns:
            Posterior samples p(x|y), of shape:
                `(batch_size, n_samples, *input_shape)`.

        """
        shape = ops.shape(measurements)

        def _tile_with_sample_dim(tensor):
            """Tile the tensor with an additional sample dimension."""
            shape = ops.shape(tensor)
            tensor = ops.repeat(tensor[:, None], n_samples, axis=1)  # (batch, n_samples, ...)
            return ops.reshape(tensor, (-1, *shape[1:]))

        measurements = _tile_with_sample_dim(measurements)
        if initial_samples is not None:
            initial_samples = ops.reshape(initial_samples, (-1, *self.input_shape))
        if "mask" in kwargs:
            kwargs["mask"] = _tile_with_sample_dim(kwargs["mask"])

        seed1, seed2 = split_seed(seed, 2)

        initial_noise = keras.random.normal(
            shape=ops.shape(measurements),
            seed=seed1,
        )

        out = self.reverse_conditional_diffusion(
            measurements=measurements,
            initial_noise=initial_noise,
            diffusion_steps=n_steps,
            initial_samples=initial_samples,
            initial_step=initial_step,
            seed=seed2,
            **kwargs,
        )
        # returns: (batch_size, n_samples, *input_shape)
        return ops.reshape(out, (shape[0], n_samples, *shape[1:]))

    def log_likelihood(self, data, **kwargs):
        """Approximate log-likelihood of the data under the model.

        Args:
            data: Data to compute log-likelihood for.
            **kwargs: Additional arguments.

        Returns:
            Approximate log-likelihood.
        """
        # This is a placeholder for likelihood estimation
        raise NotImplementedError("Likelihood computation for diffusion models not implemented yet")

    @property
    def metrics(self):
        """Metrics for training."""
        return [*self.noise_loss_tracker, *self.image_loss_tracker]

    def train_step(self, data):
        """Custom train step so we can call model.fit() on the diffusion model.
        Note:
            - Only implemented for the TensorFlow backend.
        """
        if tf is None:
            raise NotImplementedError(
                "DiffusionModel.train_step is only implemented for the TensorFlow backend."
            )

        # Get batch size and image shape
        batch_size, *input_shape = ops.shape(data)
        n_dims = len(input_shape)

        # Generate random noise
        noises = keras.random.normal(shape=ops.shape(data))

        # Sample uniform random diffusion times in [min_t, max_t]
        diffusion_times = keras.random.uniform(
            shape=[batch_size, *[1] * n_dims],
            minval=self.min_signal_rate,
            maxval=self.max_signal_rate,
        )
        noise_rates, signal_rates = self.diffusion_schedule(diffusion_times)

        # Mix data and noises
        noisy_data = signal_rates * data + noise_rates * noises

        with tf.GradientTape() as tape:
            pred_noises, pred_images = self.denoise(
                noisy_data, noise_rates, signal_rates, training=True
            )
            noise_loss = self.loss(noises, pred_noises)
            image_loss = self.loss(data, pred_images)

        gradients = tape.gradient(noise_loss, self.network.trainable_weights)
        self.optimizer.apply_gradients(zip(gradients, self.network.trainable_weights))

        self.noise_loss_tracker.update_state(noise_loss)
        self.image_loss_tracker.update_state(image_loss)

        return {m.name: m.result() for m in self.metrics}

    def diffusion_schedule(self, diffusion_times):
        """Cosine diffusion schedule https://arxiv.org/abs/2102.09672

        Args:
            diffusion_times: tensor with diffusion times in [0, 1]

        Returns:
            noise_rates: tensor with noise rates
            signal_rates: tensor with signal rates

            according to:
            - x_t = signal_rate * x_0 + noise_rate * noise
            - x_t = sqrt(alpha_t) * x_0 + sqrt(1 - alpha_t) * noise

            or with stochastic sampling:
            - x_t = sqrt(alpha_t) * x_0 + sqrt(1 - alpha_t - sigma_t^2) * noise + sigma_t * epsilon

            where:
            - sigma_t = sqrt((1 - alpha_t) / (1 - alpha_{t+1})) * sqrt(1 - alpha_{t+1} / alpha_t)

        Note:
            t+1 = previous time step
            t = current time step

        """  # noqa: E501
        # diffusion times -> angles
        start_angle = ops.cast(ops.arccos(self.max_signal_rate), "float32")
        end_angle = ops.cast(ops.arccos(self.min_signal_rate), "float32")

        diffusion_angles = start_angle + diffusion_times * (end_angle - start_angle)

        # angles -> signal and noise rates
        signal_rates = ops.cos(diffusion_angles)
        noise_rates = ops.sin(diffusion_angles)
        # note that their squared sum is always: sin^2(x) + cos^2(x) = 1
        return noise_rates, signal_rates

    def linear_diffusion_schedule(self, diffusion_times):
        """Create a linear diffusion schedule"""
        compute_alpha_t = lambda t: ops.prod(
            1 - diffusion_times[:t], axis=diffusion_times.shape[1:]
        )
        alphas = ops.vectorized_map(compute_alpha_t, ops.arange(len(diffusion_times)))
        signal_rates = ops.sqrt(alphas)
        noise_rates = ops.sqrt(1 - alphas)
        return signal_rates, noise_rates

    def denoise(self, noisy_images, noise_rates, signal_rates, training):
        """Predict noise component and calculate the image component using it"""

        pred_noises = self([noisy_images, noise_rates**2], training=training)
        pred_images = (noisy_images - noise_rates * pred_noises) / signal_rates

        return pred_noises, pred_images

    def reverse_diffusion_step(
        self,
        shape,
        pred_images,
        pred_noises,
        signal_rates,
        next_signal_rates,
        next_noise_rates,
        seed=None,
        stochastic_sampling=False,
    ):
        """A single reverse diffusion step.

        Args:
            shape: Shape of the input tensor.
            pred_images: Predicted images.
            pred_noises: Predicted noises.
            signal_rates: Current signal rates.
            next_signal_rates: Next signal rates.
            next_noise_rates: Next noise rates.
            seed: Random seed generator.
            stochastic_sampling: Whether to use stochastic sampling (DDPM).

        Returns:
            next_noisy_images: Noisy images after the reverse diffusion step.
        """
        if not stochastic_sampling:
            next_noisy_images = next_signal_rates * pred_images + next_noise_rates * pred_noises
            return next_noisy_images

        alpha_prev = signal_rates**2
        alpha = next_signal_rates**2

        sigma_t = ops.sqrt((1 - alpha) / (1 - alpha_prev)) * ops.sqrt(1 - alpha_prev / alpha)
        epsilon = keras.random.normal(shape=shape, seed=seed)

        next_noise_rates = ops.sqrt(1 - alpha - sigma_t**2)
        next_noisy_images = (
            next_signal_rates * pred_images + next_noise_rates * pred_noises + sigma_t * epsilon
        )
        return next_noisy_images

    def reverse_diffusion(
        self,
        initial_noise,
        diffusion_steps: int,
        initial_samples=None,
        initial_step: int = 0,
        stochastic_sampling: bool = False,
        seed: keras.random.SeedGenerator | None = None,
        verbose: bool = True,
        track_progress_type: Literal[None, "x_0", "x_t"] = "x_0",
    ):
        """Reverse diffusion process to generate images from noise.

        Args:
            initial_noise: Initial noise tensor.
            diffusion_steps: Number of diffusion steps.
            initial_samples: Optional initial samples to start from.
            initial_step: Initial step to start from.
            stochastic_sampling: Whether to use stochastic sampling (DDPM).
            seed: Random seed generator.
            verbose: Whether to show a progress bar.
            track_progress_type: Type of progress tracking ("x_0" or "x_t").

        Returns:
            Generated images.
        """
        num_images, *input_shape = ops.shape(initial_noise)
        step_size, progbar = self.prepare_diffusion(diffusion_steps, initial_step, verbose)

        n_dims = len(input_shape)

        base_diffusion_times = ops.ones((num_images, *[1] * n_dims)) * self.max_t

        next_noisy_images = self.prepare_schedule(
            base_diffusion_times,
            initial_noise,
            initial_samples,
            initial_step,
            step_size,
        )

        def step_fn(step, loop_state):
            noisy_images, pred_images, seed = loop_state

            # separate the current noisy image to its components
            diffusion_times = base_diffusion_times - step * step_size
            noise_rates, signal_rates = self.diffusion_schedule(diffusion_times)

            # remix the predicted components using the next signal and noise rates
            next_diffusion_times = diffusion_times - step_size
            next_noise_rates, next_signal_rates = self.diffusion_schedule(next_diffusion_times)

            # denoise
            pred_noises, pred_images = self.denoise(
                noisy_images, noise_rates, signal_rates, training=False
            )

            seed, seed1 = split_seed(seed, 2)

            next_noisy_images = self.reverse_diffusion_step(
                shape=(num_images, *input_shape),
                pred_images=pred_images,
                pred_noises=pred_noises,
                signal_rates=signal_rates,
                next_signal_rates=next_signal_rates,
                next_noise_rates=next_noise_rates,
                seed=seed1,
                stochastic_sampling=stochastic_sampling,
            )

            # this new noisy image will be used in the next step
            if progbar is not None:
                progbar.update(step + 1)

            self.store_progress(step, track_progress_type, next_noisy_images, pred_images)

            loop_state = (next_noisy_images, pred_images, seed)

            return loop_state

        _, pred_images, _ = fori_loop(
            initial_step,
            diffusion_steps,
            step_fn,
            (
                next_noisy_images,
                ops.zeros_like(initial_noise),
                seed,
            ),
            # can't jit this with progbar or tracking intermediate values
            disable_jit=verbose or track_progress_type,
        )

        return pred_images

    def reverse_conditional_diffusion(
        self,
        measurements,
        initial_noise,
        diffusion_steps: int,
        initial_samples=None,
        initial_step: int = 0,
        stochastic_sampling: bool = False,
        seed=None,
        verbose: bool = False,
        track_progress_type: Literal[None, "x_0", "x_t"] = "x_0",
        disable_jit=False,
        **kwargs,
    ):
        """Reverse diffusion process conditioned on some measurement.

        Effectively performs diffusion posterior sampling p(x_0 | y).

        Args:
            measurements: Conditioning data.
            initial_noise: Initial noise tensor.
            diffusion_steps: Number of diffusion steps.
            initial_samples: Optional initial samples to start from.
            initial_step: Initial step to start from.
            stochastic_sampling: Whether to use stochastic sampling (DDPM).
            seed: Random seed generator.
            verbose: Whether to show a progress bar.
            track_progress_type: Type of progress tracking ("x_0" or "x_t").
            **kwargs: Additional arguments. These are passed to the guidance
                function and the operator. Examples are omega, mask, etc.

        Returns:
            Generated images.

        """
        num_images, *input_shape = ops.shape(initial_noise)

        step_size, progbar = self.prepare_diffusion(
            diffusion_steps,
            initial_step,
            verbose,
        )

        n_dims = len(input_shape)
        base_diffusion_times = ops.ones((num_images, *[1] * n_dims)) * self.max_t

        next_noisy_images = self.prepare_schedule(
            base_diffusion_times,
            initial_noise,
            initial_samples,
            initial_step,
            step_size,
        )

        def step_fn(step, loop_state):
            noisy_images, pred_images, seed = loop_state

            diffusion_times = base_diffusion_times - step * step_size
            noise_rates, signal_rates = self.diffusion_schedule(diffusion_times)

            # remix the predicted components using the next signal and noise rates
            next_diffusion_times = diffusion_times - step_size
            next_noise_rates, next_signal_rates = self.diffusion_schedule(next_diffusion_times)

            gradients, (error, (pred_noises, pred_images)) = self.guidance_fn(
                noisy_images,
                measurements=measurements,
                noise_rates=noise_rates,
                signal_rates=signal_rates,
                **kwargs,
            )

            seed, seed1 = split_seed(seed, 2)
            next_noisy_images = self.reverse_diffusion_step(
                shape=(num_images, *input_shape),
                pred_images=pred_images,
                pred_noises=pred_noises,
                signal_rates=signal_rates,
                next_signal_rates=next_signal_rates,
                next_noise_rates=next_noise_rates,
                seed=seed1,
                stochastic_sampling=stochastic_sampling,
            )

            next_noisy_images = next_noisy_images - gradients
            pred_images = pred_images - gradients

            # this new noisy image will be used in the next step
            if verbose:
                progbar.update(step + 1, [("error", error)])

            self.store_progress(step, track_progress_type, next_noisy_images, pred_images)

            loop_state = (next_noisy_images, pred_images, seed)

            return loop_state

        _, pred_images, _ = fori_loop(
            initial_step,
            diffusion_steps,
            step_fn,
            (
                next_noisy_images,
                ops.zeros_like(initial_noise),
                seed,
            ),
            # can't jit this with progbar or tracking intermediate values
            disable_jit=verbose or track_progress_type or disable_jit,
        )

        return pred_images

    def prepare_diffusion(self, diffusion_steps, initial_step, verbose, disable_jit=False):
        """Prepare the diffusion process.

        This method sets up the parameters for the diffusion process, including
        validation of the initial step and calculation of the step size.
        """
        # Asserts
        if not disable_jit:
            assert initial_step >= 0, f"initial_step must be non-negative, got {initial_step}"
            assert initial_step < diffusion_steps, (
                f"initial_step must be less than diffusion_steps, got {initial_step}"
            )

        step_size = self.max_t / diffusion_steps

        if verbose:
            progbar = keras.utils.Progbar(diffusion_steps, verbose=verbose)
        else:
            progbar = None

        self.start_track_progress(diffusion_steps)

        return step_size, progbar

    def prepare_schedule(
        self,
        base_diffusion_times,
        initial_noise,
        initial_samples,
        initial_step,
        step_size,
    ):
        """Prepare the diffusion schedule.

        This method sets up the initial noisy images based on the provided
        initial noise and samples. It handles the case where the initial step
        is greater than 0, allowing for the use of partially noised images for
        initialization of the diffusion process.

        Args:
            base_diffusion_times: Base diffusion times.
            initial_noise: Initial noise tensor.
            initial_samples: Optional initial samples to start from.
            initial_step: Initial step to start from.
            step_size: Step size for the diffusion process.

        Returns:
            next_noisy_images: Noisy images after the initial step.
        """
        # We can optionally start with a set of samples that are partially noised
        if initial_samples is not None and initial_step > 0:
            starting_diffusion_times = base_diffusion_times - ((initial_step - 1) * step_size)
            noise_rates, signal_rates = self.diffusion_schedule(starting_diffusion_times)
            next_noisy_images = signal_rates * initial_samples + noise_rates * initial_noise
        elif initial_samples is not None:
            noise_rates, signal_rates = self.diffusion_schedule(base_diffusion_times)
            next_noisy_images = signal_rates * initial_samples + noise_rates * initial_noise
        elif initial_samples is None and initial_step == 0:
            # important line:
            # at the first sampling step, the "noisy image" is pure noise
            # but its signal rate is assumed to be nonzero (min_signal_rate)
            next_noisy_images = initial_noise
        else:
            raise ValueError(
                "Why are you trying to do this? Initial samples should be provided "
                "if initial_step is greater than 0 (i.e. you want to start with "
                "a partially noised image)"
            )
        return next_noisy_images

    def start_track_progress(self, diffusion_steps):
        """Initialize the progress tracking for the diffusion process.

        For diffusion animation we keep track of the diffusion progress.
        For large number of steps, we do not store all the images due to memory constraints.
        """
        self.track_progress = []
        if diffusion_steps > 50:
            self.track_progress_interval = diffusion_steps // 50
        else:
            self.track_progress_interval = 1

    def store_progress(
        self,
        step,
        track_progress_type,
        next_noisy_images,
        pred_images,
    ):
        """Store the progress of the diffusion process.

        Args:
            step: Current diffusion step.
            track_progress_type: Type of progress tracking ("x_0" or "x_t").
            next_noisy_images: Noisy images after the current step.
            pred_images: Predicted images.

        Notes:
            - x_0 is considered the predicted image (aka Tweedie estimate)
            - x_t is the noisy intermediate image
        """
        if not track_progress_type:
            return
        if step % self.track_progress_interval == 0:
            if track_progress_type == "x_0":
                self.track_progress.append(ops.convert_to_numpy(pred_images))
            elif track_progress_type == "x_t":
                self.track_progress.append(ops.convert_to_numpy(next_noisy_images))
            else:
                raise ValueError("Invalid track_progress_type")


register_presets(diffusion_model_presets, DiffusionModel)


class DiffusionGuidance(abc.ABC, Object):
    """Base class for diffusion guidance methods."""

    def __init__(self, diffusion_model, operator, disable_jit=False):
        """Initialize the diffusion guidance.

        Args:
            diffusion_model: The diffusion model to use for guidance.
            disable_jit: Whether to disable JIT compilation.
        """
        super().__init__()

        self.diffusion_model = diffusion_model
        self.operator = operator
        self.disable_jit = disable_jit
        self.setup()

    @abc.abstractmethod
    def setup(self):
        """Setup the guidance function. Should be implemented by subclasses."""
        raise NotImplementedError

    @abc.abstractmethod
    def __call__(self, *args, **kwargs):
        """Call the guidance function."""
        raise NotImplementedError


@diffusion_guidance_registry(name="dps")
class DPS(DiffusionGuidance):
    """Diffusion Posterior Sampling guidance."""

    def setup(self):
        """Setup the autograd function for DPS."""
        self.autograd = AutoGrad()
        self.autograd.set_function(self.compute_error)
        self.gradient_fn = self.autograd.get_gradient_and_value_jit_fn(
            has_aux=True,
            disable_jit=self.disable_jit,
        )

    def compute_error(
        self,
        noisy_images,
        measurements,
        noise_rates,
        signal_rates,
        omega,
        **kwargs,
    ):
        """Compute measurement error for diffusion posterior sampling.

        Args:
            noisy_images: Noisy images.
            measurement: Target measurement.
            operator: Forward operator.
            noise_rates: Current noise rates.
            signal_rates: Current signal rates.
            omega: Weight for the measurement error.
            **kwargs: Additional arguments for the operator.

        Returns:
            Tuple of (measurement_error, (pred_noises, pred_images))
        """
        pred_noises, pred_images = self.diffusion_model.denoise(
            noisy_images,
            noise_rates,
            signal_rates,
            training=False,
        )

        measurement_error = omega * L2(measurements - self.operator.forward(pred_images, **kwargs))

        return measurement_error, (pred_noises, pred_images)

    def __call__(self, noisy_images, **kwargs):
        """Call the gradient function.

        Returns a function with the following signature:
            (
                noisy_images,
                measurement,
                operator,
                noise_rates,
                signal_rates,
                omega,
                **operator_kwargs,
            ) -> gradients, (error, (pred_noises, pred_images))

        where operator_kwargs are the kwargs for the operator.

        """
        return self.gradient_fn(noisy_images, **kwargs)



"""Delay computation for ultrasound beamforming."""

import matplotlib.pyplot as plt
import numpy as np


def compute_t0_delays_planewave(probe_geometry, polar_angles, azimuth_angles=0, sound_speed=1540):
    """Computes the transmit delays for a planewave, shifted such that the
    first element fires at t=0.

    Args:
        probe_geometry (np.ndarray): The positions of the elements in the array of
            shape (n_el, 3).
        polar_angles (np.ndarray): The polar angles of the planewave in radians of shape (n_tx,).
        azimuth_angles (np.ndarray, optional): The azimuth angles of the planewave
            in radians of shape (n_tx,). Defaults to 0.
        sound_speed (float, optional): The speed of sound. Defaults to 1540.

    Returns:
        np.ndarray: The transmit delays for each element of shape (n_tx, n_el).
    """
    assert probe_geometry is not None, "Probe geometry must be provided to compute t0_delays."

    # Convert single angles to arrays for broadcasting
    polar_angles = np.atleast_1d(polar_angles)
    azimuth_angles = np.atleast_1d(azimuth_angles)

    # Compute v for all angles
    v = np.stack(
        [
            np.sin(polar_angles) * np.cos(azimuth_angles),
            np.sin(polar_angles) * np.sin(azimuth_angles),
            np.cos(polar_angles),
        ],
        axis=-1,
    )

    # Compute the projection of the element positions onto the wave vectors
    projection = np.sum(probe_geometry[:, None, :] * v, axis=-1).T

    # Convert from distance to time to compute the transmit delays.
    t0_delays_not_zero_aligned = projection / sound_speed

    # The smallest (possibly negative) time corresponds to the moment when
    # the first element fires.
    t_first_fire = np.min(t0_delays_not_zero_aligned, axis=1)

    # The transmit delays are the projection minus the offset. This ensures
    # that the first element fires at t=0.
    t0_delays = t0_delays_not_zero_aligned - t_first_fire[:, None]
    return t0_delays


def compute_t0_delays_focused(
    origins,
    focus_distances,
    probe_geometry,
    polar_angles,
    azimuth_angles=None,
    sound_speed=1540,
):
    """Computes the transmit delays for a focused transmit, shifted such that
    the first element fires at t=0.

    Args:
        origins (np.ndarray): The origin of the focused transmit of shape (n_tx, 3,).
        focus_distance (float): The distance to the focus.
        probe_geometry (np.ndarray): The positions of the elements in the array of
            shape (element, 3).
        polar_angles (np.ndarray): The polar angles of the planewave in radians of shape (n_tx,).
        azimuth_angles (np.ndarray, optional): The azimuth angles of the planewave in
            radians of shape (n_tx,).
        sound_speed (float, optional): The speed of sound. Defaults to 1540.

    Returns:
        np.ndarray: The transmit delays for each element of shape (n_tx, element).
    """
    n_tx = len(focus_distances)
    assert polar_angles.shape == (n_tx,), (
        f"polar_angles must have length n_tx = {n_tx}. Got length {len(polar_angles)}."
    )
    assert origins.shape == (n_tx, 3), (
        f"origins must have shape (n_tx, 3). Got shape {origins.shape}."
    )
    assert probe_geometry.shape[1] == 3 and probe_geometry.ndim == 2, (
        f"probe_geometry must have shape (element, 3). Got shape {probe_geometry.shape}."
    )

    # Convert single angles to arrays for broadcasting
    polar_angles = np.atleast_1d(polar_angles)
    if azimuth_angles is None:
        azimuth_angles = np.zeros(len(polar_angles))
    else:
        azimuth_angles = np.atleast_1d(azimuth_angles)
    assert azimuth_angles.shape == (n_tx,), (
        f"azimuth_angles must have length n_tx = {n_tx}. Got length {len(azimuth_angles)}."
    )

    # Compute v for all angles
    v = np.stack(
        [
            np.sin(polar_angles) * np.cos(azimuth_angles),
            np.sin(polar_angles) * np.sin(azimuth_angles),
            np.cos(polar_angles),
        ],
        axis=-1,
    )

    # Add a new dimension for broadcasting
    # The shape is now (n_tx, n_el, 3)
    v = np.expand_dims(v, axis=1)

    # Compute the location of the virtual source by adding the focus distance
    # to the origin along the wave vectors.
    virtual_sources = origins[:, None] + focus_distances[:, None, None] * v

    # Compute the distances between the virtual sources and each element
    dist = np.linalg.norm(virtual_sources - probe_geometry, axis=-1)

    # Adjust distances based on the direction of focus
    dist *= -np.sign(focus_distances[:, None])

    # Convert from distance to time to compute the
    # transmit delays/travel times.
    travel_times = dist / sound_speed

    # The smallest (possibly negative) time corresponds to the moment when
    # the first element fires.
    t_first_fire = np.min(travel_times, axis=1)

    # Shift the transmit delays such that the first element fires at t=0.
    t0_delays = travel_times - t_first_fire[:, None]

    return t0_delays


def plot_t0_delays(t0_delays):
    """Plot the t0_delays for each transducer element.

    Elements are on the x-axis, and the t0_delays are on the y-axis.
    We plot multiple lines for each angle/transmit in the scan object.

    Args:
        t0_delays (np.ndarray): The t0 delays for each element of shape (n_tx, n_el).

    """
    n_tx = t0_delays.shape[0]
    _, ax = plt.subplots()
    for tx in range(n_tx):
        ax.plot(t0_delays[tx], label=f"Transmit {tx}")
    ax.set_xlabel("Element number")
    ax.set_ylabel("t0 delay [s]")
    plt.show()



"""
zea.data.datasets
===================

This module provides classes and utilities for loading, validating, and managing
ultrasound datasets stored in HDF5 format. It supports both local and Hugging Face
Hub datasets, and offers efficient file handle caching for large collections of files.

Main Classes
------------

- H5FileHandleCache: Caches open HDF5 file handles to optimize repeated access.
- Folder: Represents a group of HDF5 files in a directory, with optional validation.
- Dataset: Provides an iterable interface over multiple HDF5 files or folders, with
    support for directory-based splitting and validation.

Functions
---------

- find_h5_files: Recursively finds HDF5 files and retrieves their dataset shapes.
- split_files_by_directory: Splits files among directories according to specified ratios.
- count_samples_per_directory: Counts the number of files per directory.

Features
--------

- Validation of dataset integrity with flag files and error logging.
- Support for Hugging Face Hub datasets with local caching.
- Utilities for dataset splitting and sample counting.
- Example usage provided in the module's main block.

"""

from collections import OrderedDict
from pathlib import Path
from typing import List

import numpy as np
import tqdm

from zea import log
from zea.data.file import File, validate_file
from zea.data.preset_utils import (
    HF_DATASETS_DIR,
    HF_PREFIX,
    _hf_list_files,
    _hf_parse_path,
    _hf_resolve_path,
)
from zea.datapaths import format_data_path
from zea.io_lib import search_file_tree
from zea.tools.hf import HFPath
from zea.utils import (
    calculate_file_hash,
    date_string_to_readable,
    get_date_string,
    reduce_to_signature,
)

_CHECK_MAX_DATASET_SIZE = 10000
_VALIDATED_FLAG_FILE = "validated.flag"
FILE_HANDLE_CACHE_CAPACITY = 128
FILE_TYPES = [".hdf5", ".h5"]


class H5FileHandleCache:
    """Cache for HDF5 file handles.

    This class manages a cache of HDF5 file handles to avoid reopening files
    multiple times. It uses an OrderedDict to maintain the order of file
    access and closes the least recently used file when the cache reaches
    its capacity."""

    def __init__(
        self,
        file_handle_cache_capacity: int = FILE_HANDLE_CACHE_CAPACITY,
    ):
        self._file_handle_cache = OrderedDict()
        self.file_handle_cache_capacity = file_handle_cache_capacity

    @staticmethod
    def _check_if_open(file):
        """Check if a file is open."""
        return bool(file.id.valid)

    def get_file(self, file_path) -> File:
        """Open an HDF5 file and cache it."""
        # If file is already in cache, return it and move it to the end
        if file_path in self._file_handle_cache:
            self._file_handle_cache.move_to_end(file_path)
            file = self._file_handle_cache[file_path]
            # if file was closed, reopen:
            if not self._check_if_open(file):
                file = File(file_path, "r")
                self._file_handle_cache[file_path] = file
        # If file is not in cache, open it and add it to the cache
        else:
            # If cache is full, close the least recently used file
            if len(self._file_handle_cache) >= self.file_handle_cache_capacity:
                _, close_file = self._file_handle_cache.popitem(last=False)
                close_file.close()
            file = File(file_path, "r")
            self._file_handle_cache[file_path] = file

        return self._file_handle_cache[file_path]

    def __del__(self):
        """Ensure cached files are closed."""
        if hasattr(self, "_file_handle_cache"):
            for _, file in self._file_handle_cache.items():
                if file is not None and self._check_if_open(file):
                    file.close()
            self._file_handle_cache = OrderedDict()


def find_h5_files(paths: str | list, key: str, search_file_tree_kwargs: dict | None = None):
    """
    Find HDF5 files from a directory or list of directories and retrieve their shapes.

    Args:
        paths (str or list): A single directory path, a list of directory paths,
            or a single HDF5 file path.
        key (str): The key to access the HDF5 dataset.
        search_file_tree_kwargs (dict, optional): Additional keyword arguments for the
            search_file_tree function. Defaults to None.

    Returns:
        - file_paths (list): List of file paths to the HDF5 files.
        - file_shapes (list): List of shapes of the HDF5 datasets.
    """

    if search_file_tree_kwargs is None:
        search_file_tree_kwargs = {}

    # Make sure paths is a list
    if not isinstance(paths, (tuple, list)):
        paths = [paths]

    file_shapes = []
    file_paths = []
    for path in paths:
        if isinstance(path, (str, Path, HFPath)) and str(path).startswith(HF_PREFIX):
            # Let File handle HF path resolution
            resolved = _hf_resolve_path(str(path))
            path = resolved

        if Path(path).is_file():
            path = Path(path)
            # If the path is a file, get its shape directly
            file_shapes.append(File.get_shape(path, key))
            file_paths.append(str(path))
            continue

        dataset_info = search_file_tree(
            path,
            filetypes=FILE_TYPES,
            hdf5_key_for_length=key,
            **search_file_tree_kwargs,
        )
        file_shapes += dataset_info["file_shapes"]
        file_paths += [str(Path(path) / file_path) for file_path in dataset_info["file_paths"]]

    return file_paths, file_shapes


class Folder:
    """Group of HDF5 files in a folder that can be validated.
    Mostly used internally, you might want to use the Dataset class instead.
    """

    def __init__(
        self,
        folder_path: list[str] | list[Path],
        key: str,
        search_file_tree_kwargs: dict | None = None,
        validate: bool = True,
        hf_cache_dir: str = HF_DATASETS_DIR,
        **kwargs,
    ):
        # Hugging Face support
        if isinstance(folder_path, (str, Path, HFPath)):
            folder_path_str = str(folder_path)
            if folder_path_str.startswith(HF_PREFIX):
                # Only resolve to local dir if it's a directory, else let File handle it
                repo_id, subpath = _hf_parse_path(folder_path_str)
                files = _hf_list_files(repo_id)
                if subpath and any(f == subpath for f in files):
                    # It's a file, let File handle it
                    pass
                else:
                    folder_path = _hf_resolve_path(folder_path_str, cache_dir=hf_cache_dir)

        super().__init__(**kwargs)

        self.folder_path = folder_path
        self.key = key
        self.search_file_tree_kwargs = search_file_tree_kwargs
        self.validate = validate
        self.file_paths, self.file_shapes = find_h5_files(
            folder_path, self.key, self.search_file_tree_kwargs
        )
        assert self.n_files > 0, f"No files in folder: {folder_path}"
        if self.validate:
            self.validate_folder()

    def __len__(self):
        """Returns the number of files in the dataset."""
        return self.n_files

    @property
    def n_files(self):
        """Return number of files in dataset."""
        return len(self.file_paths)

    def validate_folder(self):
        """Validate dataset contents.

        If a validation file exists, it checks if the dataset was validated on the same date.
        If the validation file was corrupted, it raises an error.
        If the validation file was not corrupted and validated, it prints a message and returns.
        """

        validation_file_path = self.folder_path / _VALIDATED_FLAG_FILE
        # for error logging
        validation_error_file_path = Path(
            self.folder_path, get_date_string() + "_validation_errors.log"
        )
        validation_error_log = []

        if validation_file_path.is_file():
            self._assert_validation_file(validation_file_path)
            return

        if self.n_files > _CHECK_MAX_DATASET_SIZE:
            log.warning(
                "Checking dataset in more than "
                f"{_CHECK_MAX_DATASET_SIZE} files takes too long. "
                f"Found {self.n_files} files in dataset. "
            )
            return

        num_frames_per_file = []
        validated_succesfully = True
        for file_path in tqdm.tqdm(
            self.file_paths,
            total=self.n_files,
            desc="Checking dataset files on validity (zea format)",
        ):
            try:
                validate_file(file_path)
            except Exception as e:
                validation_error_log.append(f"File {file_path} is not a valid zea dataset.\n{e}\n")
                # convert into warning
                log.warning(f"Error in file {file_path}.\n{e}")
                validated_succesfully = False

        if not validated_succesfully:
            log.warning(
                "Check warnings above for details. No validation file was created. "
                f"See {validation_error_file_path} for details."
            )
            try:
                with open(validation_error_file_path, "w", encoding="utf-8") as f:
                    for error in validation_error_log:
                        f.write(error)
            except Exception as e:
                log.error(
                    f"Could not write validation errors to {validation_error_file_path}.\n{e}"
                )
            return

        # Create the validated flag file
        self._write_validation_file(self.folder_path, num_frames_per_file)
        log.info(f"{log.green('Dataset validated.')} Check {validation_file_path} for details.")

    @staticmethod
    def _assert_validation_file(validation_file_path):
        """Check if validation file exists and is valid."""
        with open(validation_file_path, "r", encoding="utf-8") as f:
            lines = f.readlines()
            try:
                validation_date = lines[1].split(": ")[1].strip()
                read_validation_file_hash = lines[-1].split(": ")[1].strip()
            except Exception as exc:
                raise ValueError(
                    log.error(
                        f"Validation file {log.yellow(validation_file_path)} is corrupted. "
                        "Remove it if you want to redo validation."
                    )
                ) from exc

            log.info(
                f"Dataset was validated on {log.green(date_string_to_readable(validation_date))}"
            )
            log.info(f"Remove {log.yellow(validation_file_path)} if you want to redo validation.")
        # check if validation file was corrupted
        validation_file_hash = calculate_file_hash(validation_file_path, omit_line_str="hash")
        assert validation_file_hash == read_validation_file_hash, log.error(
            f"Validation file {log.yellow(validation_file_path)} was corrupted.\n"
            f"Remove it if you want to redo validation.\n"
        )

    @staticmethod
    def get_data_types(file_path):
        """Get data types from file."""
        with File(file_path) as file:
            if "data" in file:
                data_types = list(file["data"].keys())
            else:
                data_types = list(file["event_0"]["data"].keys())
        return data_types

    def _write_validation_file(self, path, num_frames_per_file):
        """Write validation file."""
        validation_file_path = Path(path, _VALIDATED_FLAG_FILE)

        # Read data types from the first file
        data_types = self.get_data_types(self.file_paths[0])

        number_of_frames = sum(num_frames_per_file)
        with open(validation_file_path, "w", encoding="utf-8") as f:
            f.write(f"Dataset: {path}\n")
            f.write(f"Validated on: {get_date_string()}\n")
            f.write(f"Number of files: {self.n_files}\n")
            f.write(f"Number of frames: {number_of_frames}\n")
            f.write(f"Data types: {', '.join(data_types)}\n")
            f.write(f"{'-' * 80}\n")
            # write all file names (not entire path) with number of frames on a new line
            for file_path, num_frames in zip(self.file_paths, num_frames_per_file):
                f.write(f"{file_path.name}: {num_frames}\n")
            f.write(f"{'-' * 80}\n")

        # Write the hash of the validation file
        validation_file_hash = calculate_file_hash(validation_file_path)
        with open(validation_file_path, "a", encoding="utf-8") as f:
            # *** validation file hash *** (80 total line length)
            f.write("*** validation file hash ***\n")
            f.write(f"hash: {validation_file_hash}")

    def __repr__(self):
        return (
            f"<zea.data.datasets.Folder at 0x{id(self):x}: "
            f"{self.n_files} files, key='{self.key}', folder='{self.folder_path}'>"
        )

    def __str__(self):
        return f"Folder with {self.n_files} files in '{self.folder_path}' (key='{self.key}')"


class Dataset(H5FileHandleCache):
    """Iterate over File(s) and Folder(s)."""

    def __init__(
        self,
        file_paths: List[str] | str,
        key: str,
        search_file_tree_kwargs: dict | None = None,
        validate: bool = True,
        directory_splits: list | None = None,
        **kwargs,
    ):
        """Initializes the Dataset.

        Args:
            file_paths (str or list): (list of) path(s) to the folder(s) containing the HDF5 file(s)
                or list of HDF5 file paths. Can be a mixed list of folders and files.
            key (str): The key to access the HDF5 dataset.
            search_file_tree_kwargs (dict, optional): Additional keyword arguments for the
                search_file_tree function. These are only used when `file_paths` are directories.
                Defaults to None.
            validate (bool, optional): Whether to validate the dataset. Defaults to True.
            directory_splits (list, optional): List of directory split by. Is a list of floats
                between 0 and 1, with the same length as the number of file_paths given.
                If none, all files in file_paths are used.

        """
        super().__init__(**kwargs)
        self.key = key
        self.search_file_tree_kwargs = search_file_tree_kwargs
        self.validate = validate

        self.file_paths, self.file_shapes = self.find_files_and_shapes(file_paths)

        if directory_splits is not None:
            # Split the files according to their parent directories
            self.file_paths, self.file_shapes = split_files_by_directory(
                self.file_paths,
                self.file_shapes,
                directory_list=file_paths,
                directory_splits=directory_splits,
            )

        assert self.n_files > 0, f"No files in file_paths: {file_paths}"

    def find_files_and_shapes(self, paths):
        """Find files and shapes in the dataset."""
        # Initialize file paths and shapes
        file_paths = []
        file_shapes = []

        if not isinstance(paths, (list, tuple)):
            paths = [paths]

        for file_path in paths:
            if isinstance(file_path, (list, tuple)):
                # If the path is a list, recursively call find_files_and_shapes
                _file_paths, _file_shapes = self.find_files_and_shapes(file_path)
                file_paths += _file_paths
                file_shapes += _file_shapes
                continue

            file_path = str(file_path)
            if file_path.startswith(HF_PREFIX):
                file_path = HFPath(file_path)
            else:
                file_path = Path(file_path)

            if file_path.is_dir():
                folder = Folder(file_path, self.key, self.search_file_tree_kwargs, self.validate)
                file_paths += folder.file_paths
                file_shapes += folder.file_shapes
                del folder
            elif file_path.is_file():
                file_paths.append(file_path)
                with File(file_path) as file:
                    file_shapes.append(file.shape(self.key))
                    if self.validate:
                        file.validate()

        return file_paths, file_shapes

    @classmethod
    def from_config(cls, dataset_folder, dtype, user=None, **kwargs):
        """Creates a Dataset from a config file."""
        path = format_data_path(dataset_folder, user)

        if "file_path" in kwargs:
            log.warning(
                "Found 'file_path' in config, this will be ignored since a Dataset is "
                + "always multiple files."
            )

        reduced_params = reduce_to_signature(cls.__init__, kwargs)
        return cls(path, key=dtype, **reduced_params)

    def __len__(self):
        """Returns the number of files in the dataset."""
        return self.n_files

    @property
    def n_files(self):
        """Return number of files in dataset."""
        return len(self.file_paths)

    def __getitem__(self, index) -> File:
        """Retrieves an item from the dataset."""
        file = self.get_file(self.file_paths[index])
        return file

    def __iter__(self):
        """
        Generator that yields images from the hdf5 files.
        """
        for idx in range(self.n_files):
            yield self[idx]

    def __call__(self):
        return iter(self)

    @property
    def total_frames(self):
        """Return total number of frames in dataset."""
        return sum(self.get_file(file_path).n_frames for file_path in self.file_paths)

    def __repr__(self):
        return (
            f"<zea.data.datasets.Dataset at 0x{id(self):x}: {self.n_files} files, key='{self.key}'>"
        )

    def __str__(self):
        return f"Dataset with {self.n_files} files (key='{self.key}')"

    def close(self):
        """Close all cached file handles."""
        for file in self._file_handle_cache.values():
            if file is not None and file.id.valid:
                file.close()
        self._file_handle_cache.clear()
        log.info("Closed all cached file handles.")

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()


def split_files_by_directory(file_names, file_shapes, directory_list, directory_splits):
    """Split files according to their parent directories and given split ratios.

    Args:
        file_names (list): List of file paths.
        file_shapes (list): List of shapes for each file.
        directory_list (list): List of directory paths to split by.
        directory_splits (list): List of split ratios (0-1) for each directory.

    Returns:
        tuple: (split_file_names, split_file_shapes)
    """
    if isinstance(directory_list, str):
        directory_list = [directory_list]
    if isinstance(directory_splits, (float, int)):
        directory_splits = [directory_splits]

    assert len(directory_splits) == len(directory_list), (
        "Number of directory splits must be equal to the number of directories."
    )
    assert all(0 <= split <= 1 for split in directory_splits), (
        "Directory splits must be between 0 and 1."
    )

    # Get directory sizes using the new count function implementation
    directory_counts = count_samples_per_directory(file_names, directory_list)
    directory_sizes = [directory_counts[str(dir_path)] for dir_path in directory_list]

    # take percentage of the files from each directory
    split_indices = [int(split * size) for split, size in zip(directory_splits, directory_sizes)]

    # offset split indices by each total number of files
    start_datasets = [0] + list(np.cumsum(directory_sizes))
    split_indices = [
        (start_datasets[i], start_datasets[i] + split) for i, split in enumerate(split_indices)
    ]

    # split the files
    split_file_names = []
    split_file_shapes = []

    for start, end in split_indices:
        split_file_names.extend(file_names[start:end])
        split_file_shapes.extend(file_shapes[start:end])

    # verify the split size
    expected_size = sum((d * s) for d, s in zip(directory_sizes, directory_splits))
    expected_size = int(expected_size)
    assert len(split_file_names) == expected_size, (
        "Number of files in split directories does not match the expected number. "
        "Please check the directory splits."
    )

    return split_file_names, split_file_shapes


def count_samples_per_directory(file_names, directories):
    """Count number of samples per directory.

    Args:
        file_names (list): List of file paths
        directories (str or list): Directory or list of directories

    Returns:
        dict: Dictionary with directory paths as keys and sample counts as values
    """
    if not isinstance(directories, list):
        directories = [directories]

    # Convert all paths to strings with normalized separators
    dir_paths = [str(Path(d)) for d in directories]

    file_paths = [str(Path(f)) for f in file_names]

    # Count files per directory using string matching
    counts = {
        dir_path: sum(1 for f in file_paths if f.startswith(dir_path)) for dir_path in dir_paths
    }

    # Assert that the total counts match the number of files
    total_count = sum(counts.values())
    assert total_count == len(file_paths), (
        f"Total count of files ({total_count}) does not match the number of files provided "
        f"({len(file_paths)}). Some files may not belong to any of the specified directories."
    )

    return counts



"""Input / output functions for reading and writing files.

Use to quickly read and write files or interact with file system.
"""

import functools
import multiprocessing
import os
import time
from io import BytesIO
from pathlib import Path

import imageio
import numpy as np
import pydicom
import tqdm
import yaml
from PIL import Image
from pydicom.pixel_data_handlers import convert_color_space

from zea import log
from zea.data.file import File

_SUPPORTED_VID_TYPES = [".avi", ".mp4", ".gif", ""]
_SUPPORTED_IMG_TYPES = [".jpg", ".png", ".JPEG", ".PNG", ".jpeg"]
_SUPPORTED_ZEA_TYPES = [".hdf5", ".h5"]


def load_video(filename):
    """Load a video file and return a numpy array of frames.

    Supported file types: avi, mp4, gif, dcm.

    Args:
        filename (str): The path to the video file.

    Returns:
        numpy.ndarray: A numpy array of frames.
    Raises:
        ValueError: If the file extension is not supported.
    """
    try:
        import cv2
    except ImportError as exc:
        raise ImportError(
            "OpenCV is required for video loading. "
            "Please install it with 'pip install opencv-python' or "
            "'pip install opencv-python-headless'."
        ) from exc

    filename = Path(filename)
    assert Path(filename).exists(), f"File {filename} does not exist"
    extension = filename.suffix
    assert extension in _SUPPORTED_VID_TYPES, f"File extension {extension} not supported"

    if extension in [".avi", ".mp4"]:
        cap = cv2.VideoCapture(filename)
        frames = []
        while True:
            ret, frame = cap.read()
            if not ret:
                break
            frames.append(frame)
        cap.release()
    elif extension == ".gif":
        frames = imageio.mimread(filename)
    elif extension == "":
        ds = pydicom.dcmread(filename)
        frames = convert_color_space(ds.pixel_array, "YBR_FULL", "RGB")
    else:
        raise ValueError("Unsupported file extension")
    frames = [cv2.cvtColor(frame, cv2.COLOR_RGB2GRAY) for frame in frames]
    return np.array(frames)


def load_image(filename, grayscale=True, color_order="RGB"):
    """Load an image file and return a numpy array.

    Supported file types: jpg, png.

    Args:
        filename (str): The path to the image file.
        grayscale (bool, optional): Whether to convert the image to grayscale. Defaults to True.
        color_order (str, optional): The desired color channel ordering. Defaults to 'RGB'.

    Returns:
        numpy.ndarray: A numpy array of the image.

    Raises:
        ValueError: If the file extension is not supported.
    """
    try:
        import cv2
    except ImportError as exc:
        raise ImportError(
            "OpenCV is required for image loading. "
            "Please install it with 'pip install opencv-python' or "
            "'pip install opencv-python-headless'."
        ) from exc

    filename = Path(filename)
    assert Path(filename).exists(), f"File {filename} does not exist"
    extension = filename.suffix
    assert extension in _SUPPORTED_IMG_TYPES, f"File extension {extension} not supported"

    image = cv2.imread(str(filename))

    if grayscale and len(image.shape) == 3:
        image = cv2.cvtColor(image, cv2.COLOR_RGB2GRAY)

    if color_order == "BGR":
        image = cv2.cvtColor(image, cv2.COLOR_RGB2BGR)
    elif color_order == "RGB":
        pass
    else:
        raise ValueError(f"Unsupported color order: {color_order}")

    return image


def search_file_tree(
    directory,
    filetypes=None,
    write=True,
    dataset_info_filename="dataset_info.yaml",
    hdf5_key_for_length=None,
    redo=False,
    parallel=False,
    verbose=True,
):
    """Lists all files in directory and sub-directories.

    If dataset_info.yaml is detected in the directory, that file is read and used
    to deduce the file paths. If not, the file paths are searched for in the
    directory and written to a dataset_info.yaml file.

    Args:
        directory (str): Path to base directory to start file search.
        filetypes (str or list, optional): Filetypes to look for in directory.
            Defaults to image types (.png etc.). Make sure to include the dot.
        write (bool, optional): Whether to write to dataset_info.yaml file.
            Defaults to True. If False, the file paths are not written to file
            and simply returned.
        dataset_info_filename (str, optional): Name of dataset info file.
            Defaults to "dataset_info.yaml", but can be changed to any name.
        hdf5_key_for_length (str, optional): Key to use for getting length of hdf5 files.
            Defaults to None. If set, the number of frames in each hdf5 file is
            calculated and stored in the dataset_info.yaml file. This is extra
            functionality of ``search_file_tree`` and only works with hdf5 files.
        redo (bool, optional): Whether to redo the search and overwrite the dataset_info.yaml file.
        parallel (bool, optional): Whether to use multiprocessing for hdf5 shape reading.
        verbose (bool, optional): Whether to print progress and info.

    Returns:
        dict: Dictionary containing file paths and total number of files.
            Has the following structure:

            .. code-block:: python

                {
                    "file_paths": list of file paths,
                    "total_num_files": total number of files,
                    "file_lengths": list of number of frames in each hdf5 file,
                    "file_shapes": list of shapes of each image file,
                    "total_num_frames": total number of frames in all hdf5 files
                }

    """
    directory = Path(directory)
    if not directory.is_dir():
        raise ValueError(
            log.error(f"Directory {directory} does not exist. Please provide a valid directory.")
        )
    assert Path(dataset_info_filename).suffix == ".yaml", (
        "Currently only YAML files are supported for dataset info file when "
        f"using `search_file_tree`, got {dataset_info_filename}"
    )

    if (directory / dataset_info_filename).is_file() and not redo:
        with open(directory / dataset_info_filename, "r", encoding="utf-8") as file:
            dataset_info = yaml.load(file, Loader=yaml.FullLoader)

        # Check if the file_shapes key is present in the dataset_info, otherwise redo the search
        if "file_shapes" in dataset_info:
            if verbose:
                log.info(
                    "Using pregenerated dataset info file: "
                    f"{log.yellow(directory / dataset_info_filename)} ..."
                )
                log.info(f"...for reading file paths in {log.yellow(directory)}")
            return dataset_info

    if redo and verbose:
        log.info(f"Overwriting dataset info file: {log.yellow(directory / dataset_info_filename)}")

    # set default file type
    if filetypes is None:
        filetypes = _SUPPORTED_IMG_TYPES + _SUPPORTED_VID_TYPES + _SUPPORTED_ZEA_TYPES

    file_paths = []

    if isinstance(filetypes, str):
        filetypes = [filetypes]

    if hdf5_key_for_length is not None:
        assert isinstance(hdf5_key_for_length, str), "hdf5_key_for_length must be a string"
        assert set(filetypes).issubset({".hdf5", ".h5"}), (
            "hdf5_key_for_length only works with when filetypes is set to "
            f"`.hdf5` or `.h5`, got {filetypes}"
        )

    # Traverse file tree to index all files from filetypes
    if verbose:
        log.info(f"Searching {log.yellow(directory)} for {filetypes} files...")
    for dirpath, _, filenames in os.walk(directory):
        for file in filenames:
            # Append to file_paths if it is a filetype file
            if Path(file).suffix in filetypes:
                file_path = Path(dirpath) / file
                file_path = file_path.relative_to(directory)
                file_paths.append(str(file_path))

    if hdf5_key_for_length is not None:
        # using multiprocessing to speed up reading hdf5 files
        # and getting the number of frames in each file
        if verbose:
            log.info("Getting number of frames in each hdf5 file...")

        get_shape_partial = functools.partial(File.get_shape, key=hdf5_key_for_length)
        # make sure to call search_file_tree from within a function
        # or use if __name__ == "__main__":
        # to avoid freezing the main process
        absolute_file_paths = [directory / file for file in file_paths]
        if parallel:
            with multiprocessing.Pool() as pool:
                file_shapes = list(
                    tqdm.tqdm(
                        pool.imap(
                            get_shape_partial,
                            absolute_file_paths,
                        ),
                        total=len(file_paths),
                        desc="Getting number of frames in each hdf5 file",
                        disable=not verbose,
                    )
                )
        else:
            file_shapes = []
            for file_path in tqdm.tqdm(
                absolute_file_paths,
                desc="Getting number of frames in each hdf5 file",
                disable=not verbose,
            ):
                file_shapes.append(File.get_shape(file_path, hdf5_key_for_length))

    assert len(file_paths) > 0, f"No image files were found in: {directory}"
    if verbose:
        log.info(f"Found {len(file_paths)} image files in {log.yellow(directory)}")
        log.info(f"Writing dataset info to {log.yellow(directory / dataset_info_filename)}")

    dataset_info = {"file_paths": file_paths, "total_num_files": len(file_paths)}
    if len(file_shapes) > 0:
        dataset_info["file_shapes"] = file_shapes
        file_lengths = [shape[0] for shape in file_shapes]
        dataset_info["file_lengths"] = file_lengths
        dataset_info["total_num_frames"] = sum(file_lengths)

    if write:
        with open(directory / dataset_info_filename, "w", encoding="utf-8") as file:
            yaml.dump(dataset_info, file)

    return dataset_info


def matplotlib_figure_to_numpy(fig):
    """Convert matplotlib figure to numpy array.

    Args:
        fig (matplotlib.figure.Figure): figure to convert.

    Returns:
        np.ndarray: numpy array of figure.

    """
    buf = BytesIO()
    fig.savefig(buf, format="png")
    buf.seek(0)
    image = Image.open(buf).convert("RGB")
    image = np.array(image)[..., :3]
    buf.close()
    return image


def retry_on_io_error(max_retries=3, initial_delay=0.5, retry_action=None):
    """Decorator to retry functions on I/O errors with exponential backoff.

    Args:
        max_retries (int): Maximum number of retry attempts.
        initial_delay (float): Initial delay between retries in seconds.
        retry_action (callable, optional): Optional function to call before each retry attempt.
            If decorating a method: ``retry_action(self, exception, attempt, *args, **kwargs)``
            If decorating a function: ``retry_action(exception, attempt, *args, **kwargs)``

    Returns:
        callable: Decorated function with retry logic.

    """

    def decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            delay = initial_delay
            last_exception = None

            for attempt in range(max_retries):
                try:
                    return func(*args, **kwargs)
                except (OSError, IOError) as e:
                    last_exception = e

                    # if args exist and first arg is a class, update retry count of that method
                    if args and hasattr(args[0], "retry_count"):
                        args[0].retry_count = attempt + 1

                    if attempt < max_retries - 1:
                        # Execute custom retry action if provided
                        if retry_action:
                            # Pass all original arguments to retry_action
                            retry_action(
                                *args,
                                exception=e,
                                retry_count=attempt,
                                **kwargs,
                            )

                        time.sleep(delay)

                    else:
                        # Last attempt failed
                        log.error(f"Failed after {max_retries} attempts: {e}")

            # If we've exhausted all retries
            raise ValueError(
                f"Failed to complete operation after {max_retries} attempts. "
                f"Last error: {last_exception}"
            )

        return wrapper

    return decorator



"""Pytorch Ultrasound Beamforming Library.

Initialize modules for registries.
"""

import numpy as np
import torch


def on_device_torch(func, inputs, device, return_numpy=False, **kwargs):
    """Applies a Torch function to inputs on a specified device.

    Args:
        func (function): Function to apply to the input data.
        inputs (ndarray): Input array.
        device (str): Device string, e.g. ``'cuda'``, ``'gpu'``, or ``'cpu'``.
        return_numpy (bool, optional): Whether to convert output
            data back to numpy. Defaults to False.
        **kwargs: Additional keyword arguments to be passed to the ``func``.

    Returns:
        torch.Tensor or ndarray: The output data.

    Raises:
        AssertionError: If ``func`` is not a function from the torch library.

    Note:
        This function converts the ``inputs`` array to a torch.Tensor and moves it to
        the specified ``device``. It then applies the ``func`` function to the inputs and
        returns the output data. If the output is a dictionary, it extracts the first value
        from the dictionary. If ``return_numpy`` is True, it converts the output data back to a
        numpy array before returning.

    Example:
        .. code-block:: python

            import torch


            def square(x):
                return x**2


            inputs = [1, 2, 3, 4, 5]
            device = "cuda"
            output = on_device_torch(square, inputs, device)
            print(output)
    """
    device = device.replace("gpu", "cuda")

    if not isinstance(inputs, torch.Tensor):
        inputs = torch.tensor(inputs)

    inputs = inputs.to(device)

    # check that function is a function from torch library
    # assert "torch" in str(type(func)), f"func: {func} should be a torch function"
    if hasattr(func, "to"):
        func = func.to(device)

    with torch.device(device):
        outputs = func(inputs, **kwargs)

    if isinstance(outputs, dict):
        # depends a bit how flexible we want to be...
        # but for now quick and dirty solution
        key = list(outputs.keys())[0]
        outputs = outputs[key]

    if return_numpy:
        if not isinstance(outputs, np.ndarray):
            outputs = outputs.cpu().numpy()

    return outputs



"""Model presets for zea.models"""

taesdxl_presets = {
    "taesdxl": {
        "metadata": {
            "description": "Tiny Autoencoder (TAESD) model",
            "params": 0,
            "path": "taesdxl",
        },
        "hf_handle": "hf://zeahub/taesdxl",
    },
}

taesdxl_encoder_presets = {
    "taesdxl_encoder": {
        "metadata": {
            "description": "Tiny encoder from TAESD model",
            "params": 0,
            "path": "taesdxl_encoder",
        },
        "hf_handle": "hf://zeahub/taesdxl",
    },
}

taesdxl_decoder_presets = {
    "taesdxl_decoder": {
        "metadata": {
            "description": "Tiny decoder from TAESD model",
            "params": 0,
            "path": "taesdxl_decoder",
        },
        "hf_handle": "hf://zeahub/taesdxl",
    },
}

echonet_dynamic_presets = {
    "echonet-dynamic": {
        "metadata": {
            "description": (
                "EchoNet-Dynamic segmentation model for cardiac ultrasound segmentation. "
                "Original paper and code: https://echonet.github.io/dynamic/"
            ),
            "params": 0,
            "path": "echonet",
        },
        "hf_handle": "hf://zeahub/echonet-dynamic",
    },
}

lpips_presets = {
    "lpips": {
        "metadata": {
            "description": "Learned Perceptual Image Patch Similarity (LPIPS) metric.",
            "params": 14716160,
            "path": "lpips",
        },
        "hf_handle": "hf://zeahub/lpips",
    },
}

unet_presets = {
    "unet-echonet-inpainter": {
        "metadata": {
            "description": (
                "U-Net model used to inpaint skipped lines (columns). "
                "Trained on 75% masked data (center values)."
            ),
            "params": 0,
            "path": "unet",
        },
        "hf_handle": "hf://zeahub/unet-echonet-inpainter",
    },
}

dense_presets = {}

diffusion_model_presets = {
    "diffusion-echonet-dynamic": {
        "metadata": {
            "description": ("Diffusion model trained on EchoNet-Dynamic dataset."),
            "params": 0,
            "path": "diffusion",
        },
        "hf_handle": "hf://zeahub/diffusion-echonet-dynamic",
    },
}

carotid_segmenter_presets = {
    "carotid-segmenter": {
        "metadata": {
            "description": (
                "Carotid segmentation model based on U-Net architecture. "
                "Trained on labeled simulated data and unlabeled invivo data."
            ),
            "params": 848461,
            "path": "carotid_segmenter",
        },
        "hf_handle": "hf://zeahub/carotid-segmenter",
    },
}



"""Structure containing parameters defining an ultrasound scan.

This module provides the :class:`Scan` class, a flexible structure
for managing all parameters related to an ultrasound scan acquisition.

Features
^^^^^^^^

- **Flexible initialization:** The :class:`Scan` class supports lazy initialization,
  allowing you to specify any combination of supported parameters. You can pass only
  the parameters you have, and the rest will be computed or set to defaults as needed.

- **Automatic computation:** Many scan properties (such as
  grid, number of pixels, wavelength, etc.) are computed automatically from the
  provided parameters. This enables you to work with minimal input and still obtain
  all necessary scan configuration details.

- **Dependency tracking and lazy evaluation:** Derived properties are computed only
  when accessed, and are automatically invalidated and recomputed if their dependencies
  change. This ensures efficient memory usage and avoids unnecessary computations.

- **Parameter validation:** All parameters are type-checked and validated against
  a predefined schema, reducing errors and improving robustness.

- **Selection of transmits:** The scan supports flexible selection of transmit events,
  using the :meth:`set_transmits` method. You can select all, a specific number,
  or specific transmit indices. The selection is stored and can be accessed via
  the :attr:`selected_transmits` property.

Comparison to ``zea.Config`` and ``zea.Probe``
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

- :class:`zea.config.Config`: A general-purpose parameter dictionary for experiment and pipeline
  configuration. It is not specific to ultrasound acquisition and does not compute
  derived parameters.

- :class:`zea.probes.Probe`: Contains only probe-specific parameters (e.g., geometry, frequency).

- :class:`zea.scan.Scan`: Combines all parameters relevant to an ultrasound acquisition,
  including probe, acquisition, and scan region. It also provides automatic computation
  of derived properties and dependency management.

Example Usage
^^^^^^^^^^^^^

.. code-block:: python

    from zea import Config, Probe, Scan

    # Initialize Scan from a Probe's parameters
    probe = Probe.from_name("verasonics_l11_4v")
    scan = Scan(**probe.get_parameters(), Nz=256)

    # Or initialize from a Config object
    config = Config.from_hf("zeahub/configs", "config_picmus_rf.yaml", repo_type="dataset")
    scan = Scan(**config.scan, n_tx=11)

    # Or manually specify parameters
    scan = Scan(
        Nx=128,
        Nz=256,
        xlims=(-0.02, 0.02),
        zlims=(0.0, 0.06),
        center_frequency=6.25e6,
        sound_speed=1540.0,
        sampling_frequency=25e6,
        n_el=128,
        n_tx=11,
    )

    # Access a derived property (computed lazily)
    grid = scan.grid  # shape: (Nz, Nx, 3)

    # Select a subset of transmit events
    scan.set_transmits(3)  # Use 3 evenly spaced transmits
    scan.set_transmits([0, 2, 4])  # Use specific transmit indices
    scan.set_transmits("all")  # Use all transmits

"""

import numpy as np
from keras import ops

from zea import log
from zea.beamform.pfield import compute_pfield
from zea.beamform.pixelgrid import check_for_aliasing, get_grid
from zea.display import compute_scan_convert_2d_coordinates, compute_scan_convert_3d_coordinates
from zea.internal.parameters import Parameters, cache_with_dependencies


class Scan(Parameters):
    """Represents an ultrasound scan configuration with computed properties.

    Args:
        Nx (int): Number of samples in the x-direction (lateral).
        Nz (int): Number of samples in the z-direction (axial).
        sound_speed (float, optional): Speed of sound in the medium in m/s.
            Defaults to 1540.0.
        sampling_frequency (float): Sampling frequency in Hz.
        center_frequency (float): Center frequency of the transducer in Hz.
        n_el (int): Number of elements in the transducer array.
        n_tx (int): Number of transmit events in the dataset.
        n_ax (int): Number of axial samples in the received signal.
        n_ch (int, optional): Number of channels (1 for RF, 2 for IQ data).
            Defaults to 1.
        xlims (tuple of float): Lateral (x) limits of the imaging region in
            meters (min, max).
        ylims (tuple of float, optional): Elevation (y) limits of the imaging
            region in meters (min, max).
        zlims (tuple of float): Axial (z) limits of the imaging region
            in meters (min, max).
        probe_geometry (np.ndarray): Element positions as array of shape (n_el, 3).
        polar_angles (np.ndarray): Polar angles for each transmit event in radians.
        azimuth_angles (np.ndarray): Azimuth angles for each transmit event in radians.
        t0_delays (np.ndarray): Transmit delays in seconds for each event.
        tx_apodizations (np.ndarray): Transmit apodizations for each event.
        focus_distances (np.ndarray): Focus distances in meters for each event.
        initial_times (np.ndarray): Initial times in seconds for each event.
        bandwidth_percent (float, optional): Bandwidth as percentage of center
            frequency. Defaults to 200.0.
        demodulation_frequency (float, optional): Demodulation frequency in Hz.
        time_to_next_transmit (np.ndarray): Time between transmit events.
        pixels_per_wavelength (int, optional): Number of pixels per wavelength.
            Defaults to 4.
        downsample (int, optional): Downsampling factor for the data. Defaults to 1.
        element_width (float, optional): Width of each transducer element in meters.
            Defaults to 0.2e-3.
        resolution (float, optional): Desired spatial resolution in meters.
        pfield_kwargs (dict, optional): Additional parameters for pressure field computation.
            See `zea.beamform.pfield.compute_pfield` for details.
        apply_lens_correction (bool, optional): Whether to apply lens correction to
            delays. Defaults to False.
        lens_thickness (float, optional): Thickness of the lens in meters.
            Defaults to None.
        f_number (float, optional): F-number of the transducer. Defaults to 1.0.
        theta_range (tuple, optional): Range of theta angles for 3D imaging.
        phi_range (tuple, optional): Range of phi angles for 3D imaging.
        rho_range (tuple, optional): Range of rho (radial) distances for 3D imaging.
        fill_value (float, optional): Value to use for out-of-bounds pixels.
            Defaults to 0.0.
        attenuation_coef (float, optional): Attenuation coefficient in dB/(MHz*cm).
            Defaults to 0.0.
        selected_transmits (None, str, int, list, or np.ndarray, optional):
            Specifies which transmit events to select.
            - None or "all": Use all transmits.
            - "center": Use only the center transmit.
            - int: Select this many evenly spaced transmits.
            - list/array: Use these specific transmit indices.
    """

    VALID_PARAMS = {
        # beamforming related parameters
        "Nx": {"type": int, "default": None},
        "Nz": {"type": int, "default": None},
        "xlims": {"type": (tuple, list), "default": None},
        "ylims": {"type": (tuple, list), "default": None},
        "zlims": {"type": (tuple, list), "default": None},
        "pixels_per_wavelength": {"type": int, "default": 4},
        "downsample": {"type": int, "default": 1},
        "resolution": {"type": float, "default": None},
        "pfield_kwargs": {"type": dict, "default": {}},
        "apply_lens_correction": {"type": bool, "default": False},
        "lens_sound_speed": {"type": (float, int), "default": None},
        "lens_thickness": {"type": float, "default": None},
        # acquisition parameters
        "sound_speed": {"type": (float, int), "default": 1540.0},
        "sampling_frequency": {"type": float, "default": None},
        "center_frequency": {"type": float, "default": None},
        "n_el": {"type": int, "default": None},
        "n_tx": {"type": int, "default": None},
        "n_ax": {"type": int, "default": None},
        "n_ch": {"type": int, "default": None},
        "bandwidth_percent": {"type": float, "default": 200.0},
        "demodulation_frequency": {"type": float, "default": None},
        "element_width": {"type": float, "default": 0.2e-3},
        "attenuation_coef": {"type": float, "default": 0.0},
        "f_number": {"type": float, "default": 1.0},
        # array parameters
        "probe_geometry": {"type": np.ndarray, "default": None},
        "polar_angles": {"type": np.ndarray, "default": None},
        "azimuth_angles": {"type": np.ndarray, "default": None},
        "t0_delays": {"type": np.ndarray, "default": None},
        "tx_apodizations": {"type": np.ndarray, "default": None},
        "focus_distances": {"type": np.ndarray, "default": None},
        "initial_times": {"type": np.ndarray, "default": None},
        "time_to_next_transmit": {"type": np.ndarray, "default": None},
        # scan conversion parameters
        "theta_range": {"type": (tuple, list), "default": None},
        "phi_range": {"type": (tuple, list), "default": None},
        "rho_range": {"type": (tuple, list), "default": None},
        "fill_value": {"type": float, "default": 0.0},
    }

    def __init__(self, **kwargs):
        # Store the current selection state before initialization
        selected_transmits_input = kwargs.pop("selected_transmits", None)

        # Initialize parent class
        super().__init__(**kwargs)

        # Initialize selection to None
        self._selected_transmits = None

        # Apply selection from input if provided
        if selected_transmits_input is not None:
            self.set_transmits(selected_transmits_input)

    @cache_with_dependencies(
        "xlims",
        "zlims",
        "Nx",
        "Nz",
        "sound_speed",
        "center_frequency",
        "pixels_per_wavelength",
    )
    def grid(self):
        """The beamforming grid of shape (Nz, Nx, 3)."""
        return get_grid(
            self.xlims,
            self.zlims,
            self.Nx,
            self.Nz,
            self.sound_speed,
            self.center_frequency,
            self.pixels_per_wavelength,
        )

    @cache_with_dependencies(
        "xlims",
        "wavelength",
    )
    def Nx(self):
        """Number of lateral (x) pixels, set to prevent aliasing if not provided."""
        Nx = self._params.get("Nx")
        if Nx is not None:
            return Nx

        width = self.xlims[1] - self.xlims[0]
        min_Nx = int(np.ceil(width / (self.wavelength / 2)))
        return max(min_Nx, 1)

    @cache_with_dependencies(
        "zlims",
        "wavelength",
    )
    def Nz(self):
        """Number of axial (z) pixels, set to prevent aliasing if not provided."""
        Nz = self._params.get("Nz")
        if Nz is not None:
            return Nz

        depth = self.zlims[1] - self.zlims[0]
        min_Nz = int(np.ceil(depth / (self.wavelength / 2)))
        return max(min_Nz, 1)

    @cache_with_dependencies("sound_speed", "center_frequency")
    def wavelength(self):
        """Calculate the wavelength based on sound speed and center frequency."""
        return self.sound_speed / self.center_frequency

    @cache_with_dependencies("sound_speed", "sampling_frequency", "n_ax")
    def z_axis(self):
        """The z-axis of the beamforming grid [m]."""
        if self.zlims is None:
            zlims = [0, self.sound_speed * self.n_ax / self.sampling_frequency / 2]
        else:
            zlims = self.zlims
        return np.linspace(zlims[0], zlims[1], self.n_ax)

    @cache_with_dependencies("grid")
    def flatgrid(self):
        """The beamforming grid of shape (Nz*Nx, 3)."""
        return self.grid.reshape(-1, 3)

    @cache_with_dependencies("n_tx")
    def selected_transmits(self):
        """Get the currently selected transmit indices.

        Returns:
            list: The list of selected transmit indices. If none were explicitly
            selected and n_tx is available, all transmits are used.
        """
        # Return all transmits if none explicitly selected
        if self._selected_transmits is None:
            if "n_tx" in self._params:
                return list(range(self._params["n_tx"]))
            return []
        return self._selected_transmits

    @cache_with_dependencies("n_tx")
    def n_tx_total(self):
        """The total number of transmits in the full dataset."""
        return self._params["n_tx"]

    @cache_with_dependencies("selected_transmits")
    def n_tx(self):
        """The number of currently selected transmits."""
        return len(self.selected_transmits)

    def _invalidate_selected_transmits(self):
        """Explicitly invalidate the selected_transmits cache and its dependents."""
        for key in [
            "selected_transmits",
            # also invalidate properties that depend on selected_transmits
            "polar_angles",
            "azimuth_angles",
            "t0_delays",
            "tx_apodizations",
            "focus_distances",
            "initial_times",
            "time_to_next_transmit",
            "pfield",
            "flat_pfield",
        ]:
            if key in self._cache:
                self._cache.pop(key)
                self._computed.discard(key)
                self._dependency_versions.pop(key, None)

    def set_transmits(self, selection):
        """Select which transmit events to use.

        This method provides flexible ways to select transmit events:

        Args:
            selection: Specifies which transmits to select:
                - None: Use all transmits
                - "all": Use all transmits
                - "center": Use only the center transmit
                - int: Select this many evenly spaced transmits
                - list/array: Use these specific transmit indices

        Returns:
            The current instance for method chaining.

        Raises:
            ValueError: If the selection is invalid or incompatible with the scan.
        """
        n_tx_total = self._params.get("n_tx")
        if n_tx_total is None:
            raise ValueError("n_tx must be set before calling set_transmits")

        # Handle None and "all" - use all transmits
        if selection is None or selection == "all":
            self._selected_transmits = None
            self._invalidate_selected_transmits()
            return self

        # Handle "center" - use center transmit
        if selection == "center":
            self._selected_transmits = [n_tx_total // 2]
            self._invalidate_selected_transmits()
            return self

        # Handle integer - select evenly spaced transmits
        if isinstance(selection, (int, np.integer)):
            selection = int(selection)  # Convert numpy integer to Python int
            if selection <= 0:
                raise ValueError(f"Number of transmits must be positive, got {selection}")

            if selection > n_tx_total:
                raise ValueError(
                    f"Requested {selection} transmits exceeds available transmits ({n_tx_total})"
                )

            if selection == 1:
                self._selected_transmits = [n_tx_total // 2]
            else:
                # Compute evenly spaced indices
                tx_indices = np.linspace(0, n_tx_total - 1, selection)
                self._selected_transmits = list(np.rint(tx_indices).astype(int))

            self._invalidate_selected_transmits()
            return self

        # Handle array-like - convert to list of indices
        if isinstance(selection, np.ndarray):
            if len(selection.shape) == 0:
                # Handle scalar numpy array
                return self.set_transmits(int(selection))
            elif len(selection.shape) == 1:
                selection = selection.tolist()
            else:
                raise ValueError(f"Invalid array shape: {selection.shape}")

        # Handle list of indices
        if isinstance(selection, list):
            # Validate indices
            if not all(isinstance(i, (int, np.integer)) for i in selection):
                raise ValueError("All transmit indices must be integers")

            if any(i < 0 or i >= n_tx_total for i in selection):
                raise ValueError(f"Transmit indices must be between 0 and {n_tx_total - 1}")

            self._selected_transmits = [
                int(i) for i in selection
            ]  # Convert numpy integers to Python ints
            self._invalidate_selected_transmits()
            return self

        # Aliasing check
        check_for_aliasing(self)

        raise ValueError(f"Unsupported selection type: {type(selection)}")

    @cache_with_dependencies("n_ch", "center_frequency")
    def demodulation_frequency(self):
        """The demodulation frequency."""
        if self._params.get("demodulation_frequency") is not None:
            return self._params["demodulation_frequency"]

        # Default behavior based on n_ch
        return self.center_frequency if self.n_ch == 2 else 0.0

    @cache_with_dependencies("selected_transmits")
    def polar_angles(self):
        """The polar angles of transmits in radians."""
        value = self._params.get("polar_angles")
        if value is None:
            log.warning("No polar angles provided, using zeros")
            value = np.zeros(self.n_tx_total)

        return value[self.selected_transmits]

    @cache_with_dependencies("selected_transmits")
    def azimuth_angles(self):
        """The azimuth angles of transmits in radians."""
        value = self._params.get("azimuth_angles")
        if value is None:
            log.warning("No azimuth angles provided, using zeros")
            value = np.zeros(self.n_tx_total)

        return value[self.selected_transmits]

    @cache_with_dependencies("selected_transmits")
    def t0_delays(self):
        """The transmit delays in seconds."""
        value = self._params.get("t0_delays")
        if value is None:
            log.warning("No transmit delays provided, using zeros")
            return np.zeros((self.n_tx_total, self.n_el))

        return value[self.selected_transmits]

    @cache_with_dependencies("selected_transmits")
    def tx_apodizations(self):
        """The transmit apodizations."""
        value = self._params.get("tx_apodizations")
        if value is None:
            log.warning("No transmit apodizations provided, using ones")
            value = np.ones((self.n_tx_total, self.n_el))

        return value[self.selected_transmits]

    @cache_with_dependencies("selected_transmits")
    def focus_distances(self):
        """The focus distances in meters."""
        value = self._params.get("focus_distances")
        if value is None:
            log.warning("No focus distances provided, using zeros")
            value = np.zeros(self.n_tx_total)

        return value[self.selected_transmits]

    @cache_with_dependencies("selected_transmits")
    def initial_times(self):
        """The initial times in seconds."""
        value = self._params.get("initial_times")
        if value is None:
            log.warning("No initial times provided, using zeros")
            value = np.zeros(self.n_tx_total)

        return value[self.selected_transmits]

    @cache_with_dependencies("selected_transmits")
    def time_to_next_transmit(self):
        """Time between transmit events."""
        value = self._params.get("time_to_next_transmit")
        if value is None:
            return None

        selected = self.selected_transmits
        return value[:, selected]

    @cache_with_dependencies(
        "sound_speed",
        "center_frequency",
        "bandwidth_percent",
        "n_el",
        "probe_geometry",
        "tx_apodizations",
        "grid",
        "t0_delays",
    )
    def pfield(self):
        """Compute or return the pressure field (pfield) for weighting."""
        pfield = compute_pfield(
            sound_speed=self.sound_speed,
            center_frequency=self.center_frequency,
            bandwidth_percent=self.bandwidth_percent,
            n_el=self.n_el,
            probe_geometry=self.probe_geometry,
            tx_apodizations=self.tx_apodizations,
            grid=self.grid,
            t0_delays=self.t0_delays,
            **self.pfield_kwargs,
        )
        return ops.convert_to_numpy(pfield)

    @cache_with_dependencies("pfield")
    def flat_pfield(self):
        """Flattened pfield for weighting."""
        return self.pfield.reshape(self.n_tx, -1).swapaxes(0, 1)

    @cache_with_dependencies("rho_range", "theta_range", "resolution", "Nz", "Nx")
    def coordinates_2d(self):
        """The coordinates for scan conversion."""
        coords, _ = compute_scan_convert_2d_coordinates(
            (self.Nz, self.Nx),
            self.rho_range,
            self.theta_range,
            self.resolution,
        )
        return coords

    @cache_with_dependencies("rho_range", "theta_range", "phi_range", "resolution", "Nz", "Nx")
    def coordinates_3d(self):
        """The coordinates for scan conversion."""
        coords, _ = compute_scan_convert_3d_coordinates(
            (self.Nz, self.Nx),
            self.rho_range,
            self.theta_range,
            self.phi_range,
            self.resolution,
        )
        return coords

    @cache_with_dependencies("time_to_next_transmit")
    def frames_per_second(self):
        """The number of frames per second [Hz]. Assumes a constant frame rate.

        Frames per second computed based on time between transmits within a frame.
        Ignores time between frames (e.g. due to processing).

        Uses the time it took to do all transmits (per frame). So if you only use some portion
        of the transmits, the fps will still be calculated based on all.
        """
        if self.time_to_next_transmit is None:
            log.warning("Time to next transmit is not set, cannot compute fps")
            return None

        # Check if fps is constant
        uniq = np.unique(self.time_to_next_transmit, axis=0)  # frame axis
        if uniq.shape[0] != 1:
            log.warning("Time to next transmit is not constant")

        # Compute fps
        time = np.mean(np.sum(self.time_to_next_transmit, axis=1))
        fps = 1 / time
        return fps
